#!/usr/bin/env python

import numpy as np
try:
    import pythreejs as p3js
except ImportError:
    p3js = None

__all__ = [
           'Box',
           'Circle',
           'Cone',
           'Cube',
           'Cylinder',
           'Icosahedron',
           'Octahedron',
           'Plane',
           'Sphere',
           'Tetrahedron',
           'Torus',
           'TorusKnot',
           'Tube'
           ]


# This is a list of ColorKeywords from THREE.js
THREE_COLORKEYWORDS = ['aliceblue', 'antiquewhite', 'aqua',
                       'aquamarine', 'azure', 'beige', 'bisque',
                       'black', 'blanchedalmond', 'blue', 'blueviolet',
                       'brown', 'burlywood', 'cadetblue', 'chartreuse',
                       'chocolate', 'coral', 'cornflowerblue',
                       'cornsilk', 'crimson', 'cyan', 'darkblue',
                       'darkcyan', 'darkgoldenrod', 'darkgray',
                       'darkgreen', 'darkgrey', 'darkkhaki',
                       'darkmagenta', 'darkolivegreen', 'darkorange',
                       'darkorchid', 'darkred', 'darksalmon',
                       'darkseagreen', 'darkslateblue', 'darkslategray',
                       'darkslategrey', 'darkturquoise', 'darkviolet',
                       'deeppink', 'deepskyblue', 'dimgray', 'dimgrey',
                       'dodgerblue', 'firebrick', 'floralwhite',
                       'forestgreen', 'fuchsia', 'gainsboro',
                       'ghostwhite', 'gold', 'goldenrod', 'gray',
                       'green', 'greenyellow', 'grey', 'honeydew',
                       'hotpink', 'indianred', 'indigo', 'ivory',
                       'khaki', 'lavender', 'lavenderblush',
                       'lawngreen', 'lemonchiffon', 'lightblue',
                       'lightcoral', 'lightcyan',
                       'lightgoldenrodyellow', 'lightgray',
                       'lightgreen', 'lightgrey', 'lightpink',
                       'lightsalmon', 'lightseagreen', 'lightskyblue',
                       'lightslategray', 'lightslategrey',
                       'lightsteelblue', 'lightyellow', 'lime',
                       'limegreen', 'linen', 'magenta', 'maroon',
                       'mediumaquamarine', 'mediumblue',
                       'mediumorchid', 'mediumpurple', 'mediumseagreen',
                       'mediumslateblue', 'mediumspringgreen',
                       'mediumturquoise', 'mediumvioletred',
                       'midnightblue', 'mintcream', 'mistyrose',
                       'moccasin', 'navajowhite', 'navy', 'oldlace',
                       'olive', 'olivedrab', 'orange', 'orangered',
                       'orchid', 'palegoldenrod', 'palegreen',
                       'paleturquoise', 'palevioletred', 'papayawhip',
                       'peachpuff', 'peru', 'pink', 'plum',
                       'powderblue', 'purple', 'red', 'rosybrown',
                       'royalblue', 'saddlebrown', 'salmon',
                       'sandybrown', 'seagreen', 'seashell', 'sienna',
                       'silver', 'skyblue', 'slateblue', 'slategray',
                       'slategrey', 'snow', 'springgreen', 'steelblue',
                       'tan', 'teal', 'thistle', 'tomato', 'turquoise',
                       'violet', 'wheat', 'white', 'whitesmoke',
                       'yellow', 'yellowgreen']

MATERIALS = ["default", "checkerboard", "metal", "dirt", "foil", "water",
             "grass"]


class Shape(object):
    """Instantiates a shape. This is primarily used as a superclass for more
    specific shapes like Cube, Cylinder, Sphere etc.

    Shapes must be associated with a reference frame and a point using the
    VisualizationFrame class.

    Parameters
    ==========

    name : str, optional
        A name assigned to the shape.
    color : str, optional
        A color string from list of colors in THREE_COLORKEYWORDS

    Examples
    ========

    >>> from pydy.viz.shapes import Shape
    >>> s = Shape()
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> a = Shape(name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'

    """
    _p3js_material_attributes = {}

    def __init__(self, name='unnamed', color='grey', material="default"):

        self.name = name
        self.color = color
        self.material = material

        self.geometry_attrs = []

    def __str__(self):

        attributes = ([self.__class__.__name__,
                       self.name,
                       'color:' + self.color,
                       'material:' + self.material] +
                      sorted([attr + ':{}'.format(getattr(self, attr)) for
                              attr in self.geometry_attrs]))

        return ' '.join(['{}'] * len(attributes)).format(*attributes)

    def __repr__(self):
        return self.__class__.__name__

    @property
    def name(self):
        """Returns the name attribute of the shape."""
        return self._name

    @name.setter
    def name(self, new_name):
        """Sets the name attribute of the shape."""
        if not isinstance(new_name, str):
            raise TypeError("'name' should be a valid str object.")
        else:
            self._name = new_name

    @property
    def color(self):
        """Returns the color attribute of the shape."""
        return self._color

    @color.setter
    def color(self, new_color):
        """Sets the color attributes of the shape. This should be a valid
        three.js color keyword string."""
        if new_color not in THREE_COLORKEYWORDS:
            msg = "'color' should be a valid Three.js colors string:\n{}"
            raise ValueError(msg.format('\n'.join(THREE_COLORKEYWORDS)))
        else:
            self._color = new_color

    @property
    def material(self):
        """Returns the material attribute of the shape."""
        return self._material

    @material.setter
    def material(self, new_material):
        """Sets the material attribute of the shape, i.e. its shine,
        brightness, opacity etc.. The material should be a valid material
        from the listed MATERIALS. If a shape is attributed as "red" color,
        and "water" material, ideally it should have opacity and brightness
        properties like that of a red fluid.

        """
        if new_material.lower() not in MATERIALS:
            msg = "'material' is not valid. Choose from:\n{}"
            raise ValueError(msg.format('\n'.join(MATERIALS)))
        else:
            self._material = new_material

    def generate_dict(self, constant_map={}):
        """Returns a dictionary containing all the data associated with the
        Shape.

        Parameters
        ==========
        constant_map : dictionary
            If any of the shape's geometry are defined as SymPy expressions,
            then this dictionary should map all SymPy Symbol's found in the
            expressions to floats.

        """
        data_dict = {}
        data_dict['name'] = self.name
        data_dict['color'] = self.color
        data_dict['material'] = self.material
        data_dict['type'] = self.__repr__()
        for geom in self.geometry_attrs:
            atr = getattr(self, geom)
            try:
                data_dict[geom] = float(atr.subs(constant_map))
            except AttributeError:
                # not a SymPy expression
                data_dict[geom] = atr
            except TypeError:
                # can't convert expression to float
                raise TypeError('{} is an expression, you '.format(atr) +
                                'must provide a mapping to numerical values.')
        return data_dict

    def _p3js_mesh(self, constant_map={}):
        """Returns a PyThreeJS mesh object that corresponds to this shape."""

        if p3js is None:
            raise ImportError('pythreejs is not installed.')

        data = self.generate_dict(constant_map=constant_map)

        attrs = dict()

        for k, v in self._p3js_attribute_map.items():
            try:
                attrs[k] = data[v]
            except KeyError:
                attrs[k] = v

        try:
            geom_attr = '{}BufferGeometry'.format(self._p3js_geometry_type)
            Geometry = getattr(p3js, geom_attr)
        except AttributeError:
            geom_attr = '{}Geometry'.format(self._p3js_geometry_type)
            Geometry = getattr(p3js, geom_attr)

        geometry = Geometry(**attrs)

        # NOTE : For some reason traitlets doesn't think 'grey' is a valid HTML
        # color. This workaround should be removed, but deprecation will likely
        # be needed.
        if data['color'] == 'grey':
            color = 'gray'
        else:
            color = data['color']

        material = p3js.MeshStandardMaterial(color=color,
                                             **self._p3js_material_attributes)

        mesh = p3js.Mesh(name=data['name'], geometry=geometry,
                         material=material)

        self._mesh = mesh

        return mesh


class Box(Shape):
    """Instantiates a box of a given size.

    Parameters
    ==========
    width : float or SymPy expression
        Width of the box along the X axis.
    height : float or SymPy expression
        Height of the box along the Y axis.
    depth : float or SymPy expression
        Depth of the box along the Z axis.

    Examples
    ========

    >>> from pydy.viz.shapes import Box
    >>> s = Box(10.0, 5.0, 1.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.width
    5.0
    >>>s.height
    1.0
    >>>s.depth
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'

    """
    _p3js_geometry_type = 'Box'
    _p3js_attribute_map = {'width': 'width',
                           'height': 'height',
                           'depth': 'depth'}

    def __init__(self, width, height, depth, **kwargs):
        super(Box, self).__init__(**kwargs)
        self.geometry_attrs += ['width', 'height', 'depth']
        self.width = width
        self.height = height
        self.depth = depth


class Cube(Shape):
    """Instantiates a cube of a given size.

    Parameters
    ==========
    length : float or SymPy expression
        The length of the cube.

    Examples
    ========

    >>> from pydy.viz.shapes import Cube
    >>> s = Cube(10.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.length
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.length = 12.0
    >>> s.length
    12.0
    >>> a = Cube('my-shape2', 'red', length=10)
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.length
    10.0

    """
    _p3js_geometry_type = 'Box'
    _p3js_attribute_map = {'width': 'length',
                           'height': 'length',
                           'depth': 'length'}

    def __init__(self, length, **kwargs):
        super(Cube, self).__init__(**kwargs)
        self.geometry_attrs.append('length')
        self.length = length


class Cylinder(Shape):
    """Instantiates a cylinder with given length and radius.

    Parameters
    ==========
    length : float or SymPy expression
        Length of the cylinder along its Y axis.
    radius : float or SymPy expression
        Radius of the cylinder (of the circular cross section normal to the Y
        axis).

    Examples
    ========

    >>> from pydy.viz.shapes import Cylinder
    >>> s = Cylinder(10.0, 5.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.length
    10.0
    >>> s.radius
    5.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.length = 12.0
    >>> s.length
    12.0
    >>> s.radius = 6.0
    >>> s.radius
    6.0
    >>> a = Cylinder(10.0, 5.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.length
    10.0
    >>> a.radius
    5.0

    """
    _p3js_geometry_type = 'Cylinder'
    _p3js_attribute_map = {'radiusTop': 'radius',
                           'radiusBottom': 'radius',
                           'radialSegments': 100,
                           'height': 'length'}

    def __init__(self, length, radius, **kwargs):
        super(Cylinder, self).__init__(**kwargs)
        self.geometry_attrs += ['length', 'radius']
        self.length = length
        self.radius = radius


class Cone(Shape):
    """Instantiates a cone with given length and base radius.

    Parameters
    ==========
    length : float or SymPy expression
        The length of the cone.
    radius : float or SymPy expression
        The base radius of the cone.

    Examples
    ========

    >>> from pydy.viz.shapes import Cone
    >>> s = Cone(10.0, 5.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.length
    10.0
    >>> s.radius
    5.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.length = 12.0
    >>> s.length
    12.0
    >>> s.radius = 6.0
    >>> s.radius
    6.0
    >>> a = Cone(10.0, 5.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.length
    10.0
    >>> a.radius
    5.0

    """
    _p3js_geometry_type = 'Cylinder'
    _p3js_attribute_map = {'radiusTop': 0.0,
                           'radiusBottom': 'radius',
                           'height': 'length'}

    def __init__(self, length, radius, **kwargs):
        super(Cone, self).__init__(**kwargs)
        self.geometry_attrs += ['length', 'radius']
        self.length = length
        self.radius = radius


class Sphere(Shape):
    """Instantiates a sphere with a given radius.

    Parameters
    ==========
    radius : float or SymPy expression
        The radius of the sphere.

    Examples
    ========

    >>> from pydy.viz.shapes import Sphere
    >>> s = Sphere(10.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.radius
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> a = Sphere(10.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0

    """
    _p3js_geometry_type = 'Sphere'
    _p3js_attribute_map = {'radius': 'radius',
                           'widthSegments': 100,
                           'heightSegments': 100}

    def __init__(self, radius=10.0, **kwargs):
        super(Sphere, self).__init__(**kwargs)
        self.geometry_attrs += ['radius']
        self.radius = radius


class Circle(Sphere):
    """Instantiates a circle with a given radius.

    Parameters
    ==========
    radius : float or SymPy Expression
        The radius of the circle.

    Examples
    ========

    >>> from pydy.viz.shapes import Circle
    >>> s = Circle(10.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.radius
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> a = Circle(10.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0

    """
    _p3js_geometry_type = 'Circle'
    _p3js_attribute_map = {'radius': 'radius',
                           'segments': 100}


class Plane(Shape):
    """Instantiates a plane with a given length and width.

    Parameters
    ==========
    length : float or SymPy expression
        Length of the plane along the Y axis.
    width : float or SymPy expression
        Width of the plane along the X axis.

    Examples
    ========

    >>> from pydy.viz.shapes import Plane
    >>> s = Plane(10.0, 5.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.length
    10.0
    >>> s.width
    5.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.length = 12.0
    >>> s.length
    12.0
    >>> s.width = 6.0
    >>> s.width
    6.0
    >>> a = Plane(10.0, 5.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.length
    10.0
    >>> a.width
    5.0

    """
    _p3js_geometry_type = 'Plane'
    _p3js_attribute_map = {'width': 'width',
                           'height': 'length'}
    _p3js_material_attributes = {'side': 'DoubleSide'}

    def __init__(self, length=10.0, width=5.0, **kwargs):
        super(Plane, self).__init__(**kwargs)
        self.geometry_attrs += ['length', 'width']
        self.length = length
        self.width = width


class Tetrahedron(Sphere):
    """Instantiates a Tetrahedron inscribed in a given radius circle.

    Parameters
    ==========
    radius : float or SymPy expression
        The radius of the circum-scribing sphere of around the tetrahedron.

    Examples
    ========

    >>> from pydy.viz.shapes import Tetrahedron
    >>> s = Tetrahedron(10.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.radius
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> a = Tetrahedron(10.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0

    """
    _p3js_geometry_type = 'Tetrahedron'
    _p3js_attribute_map = {'radius': 'radius'}


class Octahedron(Sphere):
    """Instantiaties an Octahedron inscribed in a circle of the given
    radius.

    Parameters
    ==========
    radius : float or SymPy expression.
        The radius of the circum-scribing sphere around the octahedron.

    Examples
    ========

    >>> from pydy.viz.shapes import Octahedron
    >>> s = Octahedron(10.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.radius
    10.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> a = Octahedron(10.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0

    """
    _p3js_geometry_type = 'Octahedron'
    _p3js_attribute_map = {'radius': 'radius'}


class Icosahedron(Sphere):
    """Instantiates an icosahedron inscribed in a sphere of the given
    radius.

    Parameters
    ==========
    radius : float or a SymPy expression
        Radius of the circum-scribing sphere for Icosahedron

    Examples
    ========

    >>> from pydy.viz.shapes import Icosahedron
    >>> s = Icosahedron(10)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>>s.radius
    10.0
    >>>#These can be changed later too ..
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12
    >>> a = Icosahedron(10.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0

    """
    _p3js_geometry_type = 'Icosahedron'
    _p3js_attribute_map = {'radius': 'radius'}


class Torus(Shape):
    """Instantiates a torus with a given radius and section radius.

    Parameters
    ==========
    radius : float or SymPy expression
        The radius of the torus.
    tube_radius : float or SymPy expression
        The radius of the torus tube.

    Examples
    ========

    >>> from pydy.viz.shapes import Torus
    >>> s = Torus(10.0, 5.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.radius
    10.0
    >>> s.tube_radius
    5.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> s.tube_radius = 6.0
    >>> s.tube_radius
    6.0
    >>> a = Torus(10.0, 5.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0
    >>> a.tube_radius
    5.0

    """
    _p3js_geometry_type = 'Torus'
    _p3js_attribute_map = {'radius': 'radius',
                           'tube': 'tube_radius'}

    def __init__(self, radius, tube_radius, **kwargs):
        super(Torus, self).__init__(**kwargs)
        self.geometry_attrs += ['radius', 'tube_radius']
        self.radius = radius
        self.tube_radius = tube_radius

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, new_radius):
        self._radius = new_radius

    @property
    def tube_radius(self):
        return self._tube_radius

    @tube_radius.setter
    def tube_radius(self, new_tube_radius):
        self._tube_radius = new_tube_radius


class TorusKnot(Torus):
    """Instantiates a torus knot with given radius and section radius.

    Parameters
    ==========
    radius : float or SymPy expression
        The radius of the torus knot.
    tube_radius : float or SymPy expression
        The radius of the torus knot tube.

    Examples
    ========

    >>> from pydy.viz.shapes import TorusKnot
    >>> s = TorusKnot(10.0, 5.0)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.radius
    10.0
    >>> s.tube_radius
    5.0
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 12.0
    >>> s.radius
    12.0
    >>> s.tube_radius = 6.0
    >>> s.tube_radius
    6.0
    >>> a = TorusKnot(10.0, 5.0, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    10.0
    >>> a.tube_radius
    5.0

    """
    _p3js_geometry_type = 'TorusKnot'
    _p3js_attribute_map = {'radius': 'radius',
                           'tube': 'tube_radius'}


class Tube(Shape):
    """Instantiates a tube that sweeps along a path.

    Parameters
    ==========
    radius : float or SymPy expression
        The radius of the tube.
    points : array_like, shape(n, 3)
        An array of n (x, y, z) coordinates representing points that the
        tube's center line should follow.

    Examples
    ========

    >>> from pydy.viz.shapes import Tube
    >>> points = [[1.0, 2.0, 1.0], [2.0, 1.0, 1.0], [2.0, 3.0, 4.0]]
    >>> s = Tube(10.0, points)
    >>> s.name
    'unnamed'
    >>> s.color
    'grey'
    >>> s.points
    [[1.0, 2.0, 1.0], [2.0, 1.0, 1.0], [2.0, 3.0, 4.0]]
    >>> s.name = 'my-shape1'
    >>> s.name
    'my-shape1'
    >>> s.color = 'blue'
    >>> s.color
    'blue'
    >>> s.radius = 14.0
    >>> s.radius
    14.0
    >>> s.points = [[2.0, 1.0, 4.0], [1.0, 2.0, 4.0],
    ...             [2.0, 3.0, 1.0], [1.0, 1.0, 3.0]]
    >>> s.points
    [[2.0, 1.0, 4.0], [1.0, 2.0, 4.0], [2.0, 3.0, 1.0], [1.0, 1.0, 3.0]]
    >>> a = Tube(12.0, points, name='my-shape2', color='red')
    >>> a.name
    'my-shape2'
    >>> a.color
    'red'
    >>> a.radius
    12.0
    >>> a.points
    [[1.0, 2.0, 1.0], [2.0, 1.0, 1.0], [2.0, 3.0, 4.0]]

    """

    def __init__(self, radius, points, **kwargs):
        super(Tube, self).__init__(**kwargs)
        self.geometry_attrs += ['radius', 'points']
        self.radius = radius
        self.points = points

    @property
    def points(self):
        return self._points.tolist()

    @points.setter
    def points(self, new_points):
        self._points = np.asarray(new_points)



"""The System class manages the simulation (integration) of a system whose
equations are given by KanesMethod.

Many of the attributes are also properties, and can be directly modified.

Here is the procedure for using this class.

    1. specify your options either via the constructor or via the
       attributes.
    2. optionally, call ``generate_ode_function()`` if you want to customize
       how the ODE function is generated.
    3. call ``integrate()`` to simulate your system.

The simplest usage of this class is as follows. First, we need a
KanesMethod object on which we have already invoked ``kanes_equations()``::

    km = KanesMethod(...)
    km.kanes_equations(force_list, body_list)
    times = np.linspace(0, 5, 100)
    sys = System(km, times=times)
    sys.integrate()

In this case, we use defaults for the numerical values of the constants,
specified quantities, initial conditions, etc. You probably won't like
these defaults. You can also specify such values via constructor keyword
arguments or via the attributes::

    sys = System(km,
                 initial_conditions={dynamicsymbol('q1'): 0.5},
                 times=times)
    sys.constants = {symbol('m'): 5.0}
    sys.integrate()

To double-check the constants, specifieds, states and times in your problem,
look at these properties::

    sys.constants_symbols
    sys.specifieds_symbols
    sys.states
    sys.times

In this case, the System generates the numerical ode function for you
behind the scenes. If you want to customize how this function is generated,
you must call ``generate_ode_function`` on your own::

    sys = System(KM)
    sys.generate_ode_function(generator='cython')
    sys.integrate()

"""
import warnings
from itertools import repeat

import numpy as np
import sympy as sm
from sympy.physics.mechanics import dynamicsymbols
from sympy.physics.mechanics.functions import find_dynamicsymbols
from scipy.integrate import odeint

from .codegen.ode_function_generators import generate_ode_function
from .utils import PyDyFutureWarning

SYMPY_VERSION = sm.__version__


warnings.simplefilter('once', PyDyFutureWarning)


class System(object):
    """See the class's attributes for a description of the arguments to
    this constructor.

    The parameters to this constructor are all attributes of the System.
    Actually, they are properties. With the exception of ``eom_method``,
    these attributes can be modified directly at any future point.

    Parameters
    ----------
    eom_method : sympy.physics.mechanics.KanesMethod
        You must have called ``KanesMethod.kanes_equations()`` *before*
        constructing this ``System``.
    constants : dict, optional (default: all 1.0)
        This dictionary maps SymPy Symbol objects to floats.
    specifieds : dict, optional (default: all 0)
        This dictionary maps SymPy Functions of time objects, or tuples of
        them, to floats, NumPy arrays, or functions of the state and time.
    ode_solver : function, optional (default: scipy.integrate.odeint)
        This function computes the derivatives of the states.
    initial_conditions : dict, optional (default: all zero)
        This dictionary maps SymPy Functions of time objects to floats.
    times : array_like, shape(n,), optional
        An array_like object, which contains time values over which
        equations are integrated. It has to be supplied before
        System.integrate() can be called.

    """
    def __init__(self, eom_method, constants=None, specifieds=None,
                 ode_solver=None, initial_conditions=None, times=None):

        self._eom_method = eom_method

        # TODO : What if user adds symbols after constructing a System?
        self._constants_symbols = self._Kane_constant_symbols()
        self._specifieds_symbols = self._Kane_undefined_dynamicsymbols()

        if constants is None:
            self.constants = dict()
        else:
            self.constants = constants

        if specifieds is None:
            self.specifieds = dict()
        else:
            self.specifieds = specifieds

        if ode_solver is None:
            self.ode_solver = odeint
        else:
            self.ode_solver = ode_solver

        if initial_conditions is None:
            self.initial_conditions = dict()
        else:
            self.initial_conditions = initial_conditions

        if times is None:
            self._times = []
        else:
            self._times = times

        self._evaluate_ode_function = None

    @property
    def coordinates(self):
        """Returns a list of the symbolic functions of time representing the
        system's generalized coordinates."""
        return self.eom_method.q[:]

    @property
    def speeds(self):
        """Returns a list of the symbolic functions of time representing the
        system's generalized speeds."""
        return self.eom_method.u[:]

    @property
    def states(self):
        """Returns a list of the symbolic functions of time representing the
        system's states, i.e. generalized coordinates plus the generalized
        speeds. These are in the same order as used in integration (as
        passed into evaluate_ode_function) and match the order of the mass
        matrix and forcing vector.

        """
        return self.coordinates + self.speeds

    @property
    def eom_method(self):
        """This is a sympy.physics.mechanics.KanesMethod. The method used to
        generate the equations of motion. Read-only.

        """
        return self._eom_method

    @property
    def constants(self):
        """A dict that provides the numerical values for the constants in the
        problem (all non-dynamics symbols). Keys are the symbols for the
        constants, and values are floats. Constants that are not specified in
        this dict are given a default value of 1.0.

        """
        return self._constants

    @constants.setter
    def constants(self, constants):
        self._check_constants(constants)
        self._constants = constants

    @property
    def constants_symbols(self):
        """A set of the symbolic constants (not functions of time) in the
        system.
        """
        return self._constants_symbols

    def _check_constants(self, constants):
        symbols = self.constants_symbols
        for k in constants.keys():
            if k not in symbols:
                raise ValueError("Symbol {} is not a constant.".format(k))

    def _constants_padded_with_defaults(self):
        d = dict(zip(self.constants_symbols,
                     repeat(1.0, len(self.constants_symbols))))
        d.update(self.constants)
        return d

    @property
    def specifieds(self):
        """A dict that provides numerical values for the specified quantities
        in the problem (all dynamicsymbols that are not defined by the
        equations of motion). There are two possible formats. (1) is more
        flexible, but (2) is more efficient (by a factor of 3).

        (1) Keys are the symbols for the specified quantities, or a tuple of
        symbols, and values are the floats, arrays of floats, or functions that
        generate the values. If a dictionary value is a function, it must have
        the same signature as ``f(x, t)``, the ode right-hand-side function
        (see the documentation for the ``ode_solver`` attribute). You needn't
        provide values for all specified symbols. Those for which you do not
        give a value will default to 0.0.

        (2) There are two keys: 'symbols' and 'values'. The value for 'symbols'
        is an iterable of *all* the specified quantities in the order that you
        have provided them in 'values'. Values is an ndarray, whose length is
        `len(sys.specifieds_symbols)`, or a function of x and t that returns an
        ndarray (also of length `len(sys.specifieds_symbols)`). NOTE: You must
        provide values for all specified symbols. In this case, we do *not*
        provide default values.

        NOTE: If you switch formats with the same instance of System, you
        *must* call `generate_ode_function()` before calling `integrate()`
        again.

        Examples
        --------
        Here are examples for (1). Keys can be individual symbols, or a tuple
        of symbols. Length of a value must match the length of the
        corresponding key. Values can be functions that return iterables::

            sys = System(km)
            sys.specifieds = {(a, b, c): np.ones(3), d: lambda x, t: -3 * x[0]}
            sys.specifieds = {(a, b, c): lambda x, t: np.ones(3)}

         Here are examples for (2):

            sys.specifieds = {'symbols': (a, b, c, d),
                              'values': np.ones(4)}
            sys.specifieds = {'symbols': (a, b, c, d),
                              'values': lambda x, t: np.ones(4)}

        """
        return self._specifieds

    @specifieds.setter
    def specifieds(self, specifieds):
        self._check_specifieds(specifieds)
        self._specifieds = specifieds

    @property
    def specifieds_symbols(self):
        """A set of the dynamicsymbols you must specify."""
        # TODO : Eventually use a method in the KanesMethod class.
        return self._specifieds_symbols

    def _assert_is_specified_symbol(self, symbol, all_symbols):
        if symbol not in all_symbols:
            raise ValueError("Symbol {} is not a 'specified' symbol.".format(
                symbol))

    def _assert_symbol_appears_multiple_times(self, symbol, symbols_so_far):
        if symbol in symbols_so_far:
            raise ValueError("Symbol {} appears more than once.".format(
                symbol))

    def _specifieds_are_in_format_2(self, specifieds):
        keys = specifieds.keys()
        if ('symbols' in keys and 'values' in keys):
            return True
        else:
            return False

    def _check_specifieds(self, specifieds):
        symbols = self.specifieds_symbols

        symbols_so_far = list()

        if self._specifieds_are_in_format_2(specifieds):

            # The symbols must be specifieds.
            for sym in specifieds['symbols']:
                self._assert_is_specified_symbol(sym, symbols)

            # Each specified symbol can appear only once.
            for sym in specifieds['symbols']:
                self._assert_symbol_appears_multiple_times(sym, symbols_so_far)
                symbols_so_far.append(sym)

            # Must have provided all specifieds.
            for sym in self.specifieds_symbols:
                if sym not in specifieds['symbols']:
                    raise ValueError(
                        "Specified symbol {} is not provided.".format(sym))

        else:

            for k, v in specifieds.items():

                # The symbols must be specifieds.
                if isinstance(k, tuple):
                    for ki in k:
                        self._assert_is_specified_symbol(ki, symbols)
                else:
                    self._assert_is_specified_symbol(k, symbols)

                # Each specified symbol can appear only once.
                if isinstance(k, tuple):
                    for ki in k:
                        self._assert_symbol_appears_multiple_times(
                            ki, symbols_so_far)
                        symbols_so_far.append(ki)
                else:
                    self._assert_symbol_appears_multiple_times(
                        k, symbols_so_far)
                    symbols_so_far.append(k)

    def _symbol_is_in_specifieds_dict(self, symbol, specifieds_dict):
        for k in specifieds_dict.keys():
            if symbol == k or (isinstance(k, tuple) and symbol in k):
                return True
        return False

    def _specifieds_padded_with_defaults(self):
        d = dict(zip(self.specifieds_symbols,
                     repeat(0.0, len(self.specifieds_symbols))))
        d.update(self.specifieds)
        return d

    @property
    def times(self):
        """An array-like object, containing time values over which the
        equations of motion are integrated, numerically.

        The object should be in a format which the integration module to be
        used can accept.
        """
        return self._times

    @times.setter
    def times(self, new_times):
        self._times = np.asarray(new_times)
        self._check_times(self._times)

    def _check_times(self, times):
        if len(times.shape) == 0:
            raise TypeError("Times supplied should be in an array_like format.")

        if not np.all(times >= 0):
            raise ValueError("Times supplied must have positive values.")

        if not np.all(np.diff(times) >= 0):
            raise ValueError("Times supplied should be in an ascending order.")

        return True

    @property
    def ode_solver(self):
        """A function that performs forward integration. It must have the
        same signature as scipy.integrate.odeint, which is::

            x_history = ode_solver(f, x0, t, args=(args,))

        where f is a function f(x, t, args), x0 are the initial conditions,
        x_history is the state time history, x is the state, t is the time,
        and args is a keyword argument takes arguments that are then passed
        to f. The default solver is odeint.

        """
        return self._ode_solver

    @ode_solver.setter
    def ode_solver(self, ode_solver):
        if not hasattr(ode_solver, '__call__'):
            msg = "``ode_solver`` ({}) is not a function."
            raise ValueError(msg.format(ode_solver))
        self._ode_solver = ode_solver

    @property
    def initial_conditions(self):
        """Initial conditions for all states (coordinates and speeds). Keys
        are the symbols for the coordinates and speeds, and values are
        floats. Coordinates or speeds that are not specified in this dict
        are given a default value of 0.0.

        """
        return self._initial_conditions

    @initial_conditions.setter
    def initial_conditions(self, initial_conditions):
        self._check_initial_conditions(initial_conditions)
        self._initial_conditions = initial_conditions

    def _check_initial_conditions(self, initial_conditions):
        symbols = self.states
        for k in initial_conditions.keys():
            if k not in symbols:
                raise ValueError("Symbol {} is not a state.".format(k))

    def _initial_conditions_padded_with_defaults(self):
        d = dict(zip(self.states,
                     repeat(0.0, len(self.states))))
        d.update(self.initial_conditions)
        return d

    @property
    def evaluate_ode_function(self):
        """A function generated by ``generate_ode_function`` that computes
        the state derivatives:

            x' = evaluate_ode_function(x, t, args=(...))

        This function is used by the ``ode_solver``.

        """
        return self._evaluate_ode_function

    def _args_for_gen_ode_func(self):
        """Returns a tuple of arguments in the form required by
        ``pydy.codegen.ode_function_generators.generate_ode_function``.

        """

        args = (self.eom_method.forcing_full,
                self.coordinates,
                self.speeds,
                self.constants_symbols)

        return args

    def _kwargs_for_gen_ode_func(self):
        """Returns a dictionary of arguments in the form required by
        ``pydy.codegen.ode_function_generators.generage_ode_function``.

        """

        if self._specifieds_are_in_format_2(self.specifieds):
            specifieds = self.specifieds['symbols']
        else:
            specifieds = self.specifieds_symbols

        # generate_ode_func does not accept an empty tuple for the
        # specifieds, so set it to None
        if not specifieds:
            specifieds = None

        kwargs = {'mass_matrix': self.eom_method.mass_matrix_full,
                  'specifieds': specifieds}

        return kwargs

    def generate_ode_function(self, **kwargs):
        """Calls ``pydy.codegen.ode_function_generators.generate_ode_function``
        with the appropriate arguments, and sets the ``evaluate_ode_function``
        attribute to the resulting function.

        Parameters
        ----------
        kwargs
            All other kwargs are passed onto
            ``pydy.codegen.ode_function_generators.generate_ode_function()``.
            Don't specify the ``specifieds`` keyword argument though; the
            ``System`` class takes care of those.

        Returns
        -------
        evaluate_ode_function : function
            A function which evaluates the derivaties of the states.

        """

        if 'specified' in kwargs:
            kwargs.pop('specified')
            print("User supplied 'specified' kwarg was disregarded.")

        if 'specifieds' in kwargs:
            kwargs.pop('specifieds')
            print("User supplied 'specifieds' kwarg was disregarded.")

        kwargs.update(self._kwargs_for_gen_ode_func())

        self._evaluate_ode_function = generate_ode_function(
            *self._args_for_gen_ode_func(),
            **kwargs)

        return self.evaluate_ode_function

    def integrate(self, **solver_kwargs):
        """Integrates the equations ``evaluate_ode_function()`` using
        ``ode_solver``.

        It is necessary to have first generated an ode function. If you have
        not done so, we do so automatically by invoking
        ``generate_ode_function()``. However, if you want to customize how this
        function is generated (e.g., change the generator to cython), you can
        call ``generate_ode_function()`` on your own (before calling
        ``integrate()``).

        Parameters
        ----------
        **solver_kwargs
            Optional arguments that are passed on to the ``ode_solver``.

        Returns
        -------
        x_history : np.array, shape(num_integrator_time_steps, 2)
            The trajectory of states (coordinates and speeds) through the
            requested time interval. num_integrator_time_steps is either
            len(times) if len(times) > 2, or is determined by the
            ``ode_solver``.

        """
        # Users might have changed these properties by directly accessing the
        # dict, without using the setter. Before we integrate, make sure they
        # did not muck up these dicts.
        self._check_constants(self.constants)
        self._check_specifieds(self.specifieds)
        self._check_initial_conditions(self.initial_conditions)
        self._check_times(self.times)

        if self.evaluate_ode_function is None:
            self.generate_ode_function()

        init_conds_dict = self._initial_conditions_padded_with_defaults()
        initial_conditions_in_proper_order = \
            [init_conds_dict[k] for k in self.states]

        if self._specifieds_are_in_format_2(self.specifieds):
            specified_value = self.specifieds['values']
        else:
            specified_value = self._specifieds_padded_with_defaults()

        # If there are no specifieds then specified_value will be an empty
        # dict.
        if isinstance(specified_value, dict) and not specified_value:
            args = (self._constants_padded_with_defaults(),)
        else:
            args = (specified_value, self._constants_padded_with_defaults())

        x_history = self.ode_solver(
            self.evaluate_ode_function,
            initial_conditions_in_proper_order,
            self.times,
            args=args, **solver_kwargs)

        return x_history

    def _Kane_inlist_insyms(self):
        """TODO temporary."""

        uaux = self.eom_method._uaux[:]
        uauxdot = [sm.diff(i, dynamicsymbols._t) for i in uaux]

        # Checking for dynamic symbols outside the dynamic differential
        # equations; throws error if there is.

        # TODO : KanesMethod should provide public attributes for qdot,
        # udot, uaux, and uauxdot.
        # NOTE : There have been breaking changes in SymPy, e.g.
        # https://github.com/pydy/pydy/issues/395, so all of these need to be
        # converted to lists before summing.
        insyms = set(list(self.eom_method.q[:]) +
                     list(self.eom_method._qdot[:]) +
                     list(self.eom_method.u[:]) +
                     list(self.eom_method._udot[:]) +
                     list(uaux) + list(uauxdot))

        inlist = (self.eom_method.forcing_full[:] +
                  self.eom_method.mass_matrix_full[:])

        return inlist, insyms

    def _Kane_undefined_dynamicsymbols(self):
        """Similar to ``_find_dynamicsymbols()``, except that it checks all
        syms used in the system. Code is copied from ``linearize()``.

        TODO temporarily here until KanesMethod and Lagranges method have an
        interface for obtaining these quantities.

        """
        from_eoms, from_sym_lists = self._Kane_inlist_insyms()
        functions_of_time = set()
        for expr in from_eoms:
            functions_of_time = functions_of_time.union(
                find_dynamicsymbols(expr))
        return functions_of_time.difference(from_sym_lists)

    def _Kane_constant_symbols(self):
        """Similar to ``_find_othersymbols()``, except it checks all syms used in
        the system.

        Remove the time symbol.

        TODO temporary.

        """
        from_eoms, from_sym_lists = self._Kane_inlist_insyms()
        unique_symbols = set()
        for expr in from_eoms:
            unique_symbols = unique_symbols.union(expr.free_symbols)
        constants = unique_symbols
        constants.remove(dynamicsymbols._t)
        return constants



#!/usr/bin/env python

"""This module contains source code dedicated to generating C code from
matrices generated from sympy.physics.mechanics."""

import os

from pkg_resources import parse_version
import sympy as sm

SYMPY_VERSION = sm.__version__

try:
    try:
        if parse_version(SYMPY_VERSION) >= parse_version('1.7'):
            from sympy.printing.c import C99CodePrinter as CCodePrinter
        else:
            from sympy.printing.ccode import C99CodePrinter as CCodePrinter
    except ImportError:
        # SymPy 1.0 and lower uses this version.
        from sympy.printing.ccode import CCodePrinter
except ModuleNotFoundError:
    # SymPy > 1.5 renamed the module ccode to c
    from sympy.printing.c import C99CodePrinter as CCodePrinter

from .matrix_generator import MatrixGenerator
from ..utils import wrap_and_indent


class CMatrixGenerator(MatrixGenerator):
    """This class generates C source files that simultaneously numerically
    evaluate any number of SymPy matrices.

    """

    _base_printer = CCodePrinter

    _c_header_template = """\
void evaluate(
{input_args}
{output_args}
             );
/*

{input_docstring}

*/"""

    _c_source_template = """\
#include <math.h>{header_include}

void evaluate(
{input_args}
{output_args}
             )
{{

{subexprs}

{outputs}

}}\
"""

    def _generate_code_blocks(self):
        """Writes the blocks of code for the C file."""

        # TODO : This could use the super classes method with some tweaks.

        printer = self._generate_pydy_printer()()

        self.code_blocks = {}

        lines = []
        for i, input_arg in enumerate(self.arguments):
            lines.append('double input_{}[{}],'.format(i, len(input_arg)))
        self.code_blocks['input_args'] = wrap_and_indent(lines, 14)

        lines = []
        for i, output_arg in enumerate(self.matrices):
            nr, nc = output_arg.shape
            lines.append('double output_{}[{}],'.format(i, nr * nc))
            self.code_blocks['output_args'] = \
                wrap_and_indent(lines, 14)[:-1]  # remove last comma

        lines = []
        for i, (input_arg, explan) in enumerate(zip(self.arguments,
                                                    self.comma_lists())):
            lines.append('input_{}[{}] : [{}]'.format(i, len(input_arg),
                                                      explan))
        self.code_blocks['input_docstring'] = wrap_and_indent(lines, 0)

        lines = []
        for var, expr in self.subexprs:
            var_str = printer.doprint(var)
            expr_str = printer.doprint(expr)
            lines.append('double {} = {};'.format(var_str, expr_str))
        self.code_blocks['subexprs'] = wrap_and_indent(lines)

        outputs = ''
        for i, output in enumerate(self.simplified_matrices):
            nr, nc = output.shape
            lhs = sm.MatrixSymbol('output_{}'.format(i), nr, nc)
            try:
                code_str = printer.doprint(output, lhs)
            except AttributeError:
                # The above fails in SymPy 0.7.4.1 because Matrix printing
                # isn't supported.
                code_lines = []
                for j, element in enumerate(output):
                    assignment = 'output_{}[{}]'.format(i, j)
                    code_lines.append(printer.doprint(element, assignment))
                code_str = '\n'.join(code_lines)
            outputs += wrap_and_indent(code_str.split('\n'))
            if i != len(self.simplified_matrices) - 1:
                outputs += '\n\n'  # space between each output

        self.code_blocks['outputs'] = outputs

    def doprint(self, prefix=None):
        """Returns a string each for the header and the source files.

        Parameters
        ==========
        prefix : string, optional
            A prefix for the name of the header file. This will cause an
            include statement to to be added to the source.

        """

        if prefix is not None:
            filling = {'header_include': '\n#include "{}.h"'.format(prefix)}
        else:
            filling = {'header_include': ''}

        filling.update(self.code_blocks)

        c_header = self._c_header_template.format(**filling)
        c_source = self._c_source_template.format(**filling)

        return c_header, c_source

    def write(self, prefix, path=None):
        """Writes a header and source file to disk.

        Parameters
        ==========
        prefix : string
            Two files will be generated: ``<prefix>.c`` and
            ``<prefix>.h``.

        """

        if path is None:
            path = os.getcwd()

        header, source = self.doprint(prefix=prefix)

        with open(os.path.join(path, prefix + '.h'), 'w') as f:
            f.write(header)

        with open(os.path.join(path, prefix + '.c'), 'w') as f:
            f.write(source)


class _CLUsolveGenerator(CMatrixGenerator):
    """This is a private undocumented class that supports the
    ``linear_sys_solver='sympy'`` in CythonMatrixGenerator. It cse's A and b of
    a linear system Ax=b, then solves the linear system symbolically and cse's
    the result x. This is a more efficient way to get the symbolic solution of
    a linear system encoded in generated C code."""

    def _generate_cse(self, prefix='pydy_'):
        # NOTE : This assumes the first two items in self.matrices are A and b
        # of and Ax=b system. This also ignores cse=False.

        gen1 = sm.numbered_symbols(prefix)
        subexprs1, mats_simp = sm.cse(self.matrices, symbols=gen1)

        A_simp = mats_simp[0]
        b_simp = mats_simp[1]

        x = A_simp.LUsolve(b_simp)

        gen2 = sm.numbered_symbols(prefix, start=len(subexprs1))
        subexprs2, x_simp = sm.cse(x, symbols=gen2)

        # swap the b matrix with the x result
        mats_simp[1] = x_simp[0]

        self.subexprs = subexprs1 + subexprs2
        self.simplified_matrices = tuple(mats_simp)



__all__ = []

# The following pattern is used below for importing sub-modules:
#
# 1. "from foo import *".  This imports all the names from foo.__all__ into
#    this module. But, this does not put those names into the __all__ of
#    this module. This enables "from sympy.physics.mechanics import kinematics" to
#    work.
# 2. "import foo; __all__.extend(foo.__all__)". This adds all the names in
#    foo.__all__ to the __all__ of this module. The names in __all__
#    determine which names are imported when
#    "from sympy.physics.mechanics import *" is done.

from . import visualization_frame
from .visualization_frame import *
__all__.extend(visualization_frame.__all__)

from . import shapes
from .shapes import *
__all__.extend(shapes.__all__)

from . import scene
from .scene import *
__all__.extend(scene.__all__)

from . import camera
from .camera import *
__all__.extend(camera.__all__)

from . import light
from .light import *
__all__.extend(light.__all__)

from . import server
from .server import *
__all__.extend(server.__all__)



#!/usr/bin/env python

import os

try:
    from sympy.printing.octave import OctaveCodePrinter
except ImportError:
    raise ImportError('Octave code printing is only available with SymPy >= 0.7.6.')

from .matrix_generator import MatrixGenerator


class OctaveMatrixGenerator(MatrixGenerator):
    """This class generates Octave/Matlab source files that simultaneously
    numerically evaluate any number of SymPy matrices.

    """

    _idx_start = 1
    _idx_delim = "()"
    _base_printer = OctaveCodePrinter
    _type_declar = ''  # prepended to variable introductions
    _line_contin = ' ...'
    _comment_char = '%'

    # TODO : The first two lines will not wrap. For many inputs/outputs it
    # would be nice to have some wrapping.
    _m_template = """\
function [{output_args}] = {prefix}({input_args})
% function [{output_args}] = {prefix}({input_args})
%
{docstring}

{subexprs}

{outputs}

end
"""

    def doprint(self, prefix='eval_mats'):
        """Returns a string that implements the function.

        Parameters
        ==========
        prefix : string, optional
            The name of the Octave/Matlab function.

        """
        self.code_blocks['prefix'] = prefix

        return self._m_template.format(**self.code_blocks)

    def write(self, prefix='eval_mats', path=None):
        """Writes the <prefix>.m file to disc at the give path location."""

        if path is None:
            path = os.getcwd()

        text = self.doprint(prefix=prefix)

        with open(os.path.join(path, prefix + '.m'), 'w') as f:
            f.write(text)



#!/usr/bin/env python

import os

from pkg_resources import parse_version
import sympy as sm
from nose.tools import assert_raises

from ...models import multi_mass_spring_damper
from ..c_code import CMatrixGenerator

SYMPY_VERSION = sm.__version__


class TestCMatrixGenerator():

    def setup(self):

        self.prefix = 'boogly_bee'

        sys = multi_mass_spring_damper(6, True, True)

        self.matrices = (sys.eom_method.mass_matrix,
                         sys.eom_method.forcing)

        # NOTE : ordered is used here because this order is different in
        # different versions of SymPy.
        self.arguments = (list(sm.ordered(sys.constants_symbols)),
                          sys.coordinates,
                          sys.speeds,
                          list(sm.ordered(sys.specifieds_symbols)))

        self.generator = CMatrixGenerator(self.arguments, self.matrices)

    def test_init(self):

        assert self.generator.matrices == self.matrices
        assert self.generator.arguments == self.arguments

        # Make sure an error is risen if not enough arguments are provided.
        arguments = self.arguments[:-1]

        assert_raises(ValueError, CMatrixGenerator, arguments, self.matrices)

    def test_generate_cse(self):

        pd = sm.symbols('pydy_:15')

        (c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, m0, m1, m2, m3,
         m4, m5) = self.arguments[0]
        x0, x1, x2, x3, x4, x5 = self.arguments[1]
        v0, v1, v2, v3, v4, v5 = self.arguments[2]
        f0, f1, f2, f3, f4, f5 = self.arguments[3]

        if parse_version(SYMPY_VERSION) >= parse_version('1.2'):
            expected_subexprs = [
                (pd[0], m4 + m5),
                (pd[1], m3 + pd[0]),
                (pd[2], m2 + pd[1]),
                (pd[3], m1 + pd[2]),
                (pd[4], g*m5 + f5),
                (pd[5], g*m4 + pd[4] + f4),
                (pd[6], g*m3 + pd[5] + f3),
                (pd[7], g*m2 + pd[6] + f2),
                (pd[8], g*m1 + pd[7] + f1)]
            expected_simplified_matrices = (
                sm.Matrix([[m0 + pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [pd[2], pd[2], pd[2], pd[1], pd[0], m5],
                           [pd[1], pd[1], pd[1], pd[1], pd[0], m5],
                           [pd[0], pd[0], pd[0], pd[0], pd[0], m5],
                           [m5,     m5,     m5,     m5,     m5, m5]]),
                sm.Matrix([[-c0*v0 + g*m0 - k0*x0 + pd[8] + f0],
                           [-c1*v1 - k1*x1 + pd[8]],
                           [-c2*v2 - k2*x2 + pd[7]],
                           [-c3*v3 - k3*x3 + pd[6]],
                           [-c4*v4 - k4*x4 + pd[5]],
                           [-c5*v5 - k5*x5 + pd[4]]]))
        elif parse_version(SYMPY_VERSION) >= parse_version('1.1'):
            expected_subexprs = [
                (pd[0], m4 + m5),
                (pd[1], m3 + pd[0]),
                (pd[2], m2 + pd[1]),
                (pd[3], m1 + pd[2]),
                (pd[4], f2),
                (pd[5], f3),
                (pd[6], f4),
                (pd[7], f5),
                (pd[8], g*m2),
                (pd[9], g*m3),
                (pd[10], g*m4),
                (pd[11], g*m5),
                (pd[12], g*m1 + pd[10] + pd[11] + pd[4] + pd[5] + pd[6] +
                 pd[7] + pd[8] + pd[9] + f1)]
            expected_simplified_matrices = (
                sm.Matrix([[m0 + pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [pd[2], pd[2], pd[2], pd[1], pd[0], m5],
                           [pd[1], pd[1], pd[1], pd[1], pd[0], m5],
                           [pd[0], pd[0], pd[0], pd[0], pd[0], m5],
                           [m5,     m5,     m5,     m5,     m5, m5]]),
                sm.Matrix([[-c0*v0 + g*m0 - k0*x0 + pd[12] + f0],
                           [-c1*v1 - k1*x1 + pd[12]],
                           [-c2*v2 - k2*x2 + pd[10] + pd[11] + pd[4] + pd[5] +
                            pd[6] + pd[7] + pd[8] + pd[9]],
                           [-c3*v3 - k3*x3 + pd[10] + pd[11] + pd[5] + pd[6] +
                            pd[7] + pd[9]],
                           [-c4*v4 - k4*x4 + pd[10] + pd[11] + pd[6] + pd[7]],
                           [-c5*v5 - k5*x5 + pd[11] + pd[7]]]))
        elif parse_version(SYMPY_VERSION) > parse_version('1.0'):
            expected_subexprs = [
                (pd[0], m4 + m5),
                (pd[1], m3 + pd[0]),
                (pd[2], m2 + pd[1]),
                (pd[3], m1 + pd[2]),
                (pd[4], f2),
                (pd[5], f3),
                (pd[6], f4),
                (pd[7], f5),
                (pd[8], g*m2),
                (pd[9], g*m3),
                (pd[10], g*m4),
                (pd[11], g*m5),
                (pd[12], g*m1 + pd[10] + pd[11] + pd[4] + pd[5] + pd[6] +
                 pd[7] + pd[8] + pd[9] + f1),
                (pd[13], pd[10] + pd[11] + pd[5] + pd[6] + pd[7] + pd[9]),
                (pd[14], pd[11] + pd[7])]

            expected_simplified_matrices = (
                sm.Matrix([[m0 + pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [     pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                           [     pd[2], pd[2], pd[2], pd[1], pd[0], m5],
                           [     pd[1], pd[1], pd[1], pd[1], pd[0], m5],
                           [     pd[0], pd[0], pd[0], pd[0], pd[0], m5],
                           [        m5,    m5,    m5,    m5,    m5, m5]]),
               sm.Matrix([[    -c0*v0 + g*m0 - k0*x0 + pd[12] + f0],
                          [                   -c1*v1 - k1*x1 + pd[12]],
                          [ -c2*v2 - k2*x2 + pd[13] + pd[4] + pd[8]],
                          [                   -c3*v3 - k3*x3 + pd[13]],
                          [-c4*v4 - k4*x4 + pd[10] + pd[14] + pd[6]],
                          [                   -c5*v5 - k5*x5 + pd[14]]]))
        else:
            expected_subexprs = [
                (pd[0], m4 + m5),
                (pd[1], m3 + pd[0]),
                (pd[2], m2 + pd[1]),
                (pd[3], m1 + pd[2]),
                (pd[4], f2),
                (pd[5], f3),
                (pd[6], f4),
                (pd[7], f5),
                (pd[8], g*m2),
                (pd[9], g*m3),
                (pd[10], g*m4),
                (pd[11], g*m5),
                (pd[12], (g*m1 + pd[10] + pd[11] + pd[4] + pd[5] + pd[6] +
                          pd[7] + pd[8] + pd[9] + f1))]

            expected_simplified_matrices = (
                sm.Matrix([[m0 + pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                        [pd[3],      pd[3], pd[2], pd[1], pd[0], m5],
                        [pd[2],      pd[2], pd[2], pd[1], pd[0], m5],
                        [pd[1],      pd[1], pd[1], pd[1], pd[0], m5],
                        [pd[0],      pd[0], pd[0], pd[0], pd[0], m5],
                        [m5,         m5,    m5,    m5,    m5,    m5]]),
                sm.Matrix([-c0*v0 + g*m0 - k0*x0 + pd[12] + f0,
                        -c1*v1 - k1*x1 + pd[12],
                        -c2*v2 - k2*x2 + pd[10] + pd[11] + pd[4] + pd[5] +
                            pd[6] + pd[7] + pd[8] + pd[9],
                        -c3*v3 - k3*x3 + pd[10] + pd[11] + pd[5] + pd[6] +
                            pd[7] + pd[9],
                        -c4*v4 - k4*x4 + pd[10] + pd[11] + pd[6] + pd[7],
                        -c5*v5 - k5*x5 + pd[11] + pd[7]]))

        self.generator._generate_cse()

        assert self.generator.subexprs == expected_subexprs
        assert self.generator.simplified_matrices == expected_simplified_matrices

    def test_skip_cse(self):

        (c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, m0, m1, m2, m3,
         m4, m5) = self.arguments[0]
        x0, x1, x2, x3, x4, x5 = self.arguments[1]
        v0, v1, v2, v3, v4, v5 = self.arguments[2]
        f0, f1, f2, f3, f4, f5 = self.arguments[3]

        expected_subexprs = []
        pd = 13 * [0]

        pd[0] = m4 + m5
        pd[1] = m3 + pd[0]
        pd[2] = m2 + pd[1]
        pd[3] = m1 + pd[2]
        pd[4] = f2
        pd[5] = f3
        pd[6] = f4
        pd[7] = f5
        pd[8] = g*m2
        pd[9] = g*m3
        pd[10] = g*m4
        pd[11] = g*m5
        pd[12] = (g*m1 + pd[10] + pd[11] + pd[4] + pd[5] + pd[6] + pd[7] +
                  pd[8] + pd[9] + f1)

        expected_simplified_matrices = (
            sm.Matrix([[m0 + pd[3], pd[3], pd[2], pd[1], pd[0], m5],
                       [pd[3],      pd[3], pd[2], pd[1], pd[0], m5],
                       [pd[2],      pd[2], pd[2], pd[1], pd[0], m5],
                       [pd[1],      pd[1], pd[1], pd[1], pd[0], m5],
                       [pd[0],      pd[0], pd[0], pd[0], pd[0], m5],
                       [m5,         m5,    m5,    m5,    m5,    m5]]),
            sm.Matrix([-c0*v0 + g*m0 - k0*x0 + pd[12] + f0,
                       -c1*v1 - k1*x1 + pd[12],
                       -c2*v2 - k2*x2 + pd[10] + pd[11] + pd[4] + pd[5] +
                           pd[6] + pd[7] + pd[8] + pd[9],
                       -c3*v3 - k3*x3 + pd[10] + pd[11] + pd[5] + pd[6] +
                           pd[7] + pd[9],
                       -c4*v4 - k4*x4 + pd[10] + pd[11] + pd[6] + pd[7],
                       -c5*v5 - k5*x5 + pd[11] + pd[7]]))

        self.generator._ignore_cse()

        assert self.generator.subexprs == expected_subexprs
        assert self.generator.simplified_matrices == expected_simplified_matrices

    def test_generate_pydy_printer(self):

        PyDyCCodePrinter = self.generator._generate_pydy_printer()

        printer = PyDyCCodePrinter()

        assert printer.doprint(self.arguments[0][3]) == 'input_0[3]'
        assert printer.doprint(self.arguments[1][5]) == 'input_1[5]'
        assert printer.doprint(self.arguments[2][1]) == 'input_2[1]'
        assert printer.doprint(self.arguments[3][2]) == 'input_3[2]'

    def test_generate_comma_lists(self):

        expected = (('c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, '
                     'm0, m1, m2, m3, m4, m5'),
                    'x0(t), x1(t), x2(t), x3(t), x4(t), x5(t)',
                    'v0(t), v1(t), v2(t), v3(t), v4(t), v5(t)',
                    'f0(t), f1(t), f2(t), f3(t), f4(t), f5(t)')

        assert self.generator.comma_lists() == expected

    def test_generate_code_blocks(self):

        expected = {}

        expected['input_args'] = \
"""\
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],\
"""

        expected['output_args'] = \
"""\
              double output_0[36],
              double output_1[6]\
"""

        expected['input_docstring'] = \
"""\
input_0[19] : [c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, m0, m1, m2,
m3, m4, m5]
input_1[6] : [x0(t), x1(t), x2(t), x3(t), x4(t), x5(t)]
input_2[6] : [v0(t), v1(t), v2(t), v3(t), v4(t), v5(t)]
input_3[6] : [f0(t), f1(t), f2(t), f3(t), f4(t), f5(t)]\
"""
        if parse_version(SYMPY_VERSION) >= parse_version('1.2'):
            expected['subexprs'] = \
"""\
    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_0[6] * input_0[18] + input_3[5];
    double pydy_5 = input_0[6] * input_0[17] + pydy_4 + input_3[4];
    double pydy_6 = input_0[6] * input_0[16] + pydy_5 + input_3[3];
    double pydy_7 = input_0[6] * input_0[15] + pydy_6 + input_3[2];
    double pydy_8 = input_0[6] * input_0[14] + pydy_7 + input_3[1];\
"""
            expected['outputs'] = \
"""\
    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_8 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_8;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_7;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_6;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_5;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_4;\
"""
        elif parse_version(SYMPY_VERSION) >= parse_version('1.1'):
            expected['subexprs'] = \
"""\
    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];\
"""
            expected['outputs'] = \
"""\
    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_10
    + pydy_11 + pydy_4 + pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_10
    + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_11 + pydy_6 + pydy_7;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_11
    + pydy_7;\
"""
        elif parse_version(SYMPY_VERSION) > parse_version('1.0'):

            expected['subexprs'] = \
"""\
    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];
    double pydy_13 = pydy_10 + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    double pydy_14 = pydy_11 + pydy_7;\
"""
            expected['outputs'] = \
"""\
    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_13
    + pydy_4 + pydy_8;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] +
    pydy_13;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_14 + pydy_6;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] +
    pydy_14;\
"""

        else:

            expected['subexprs'] = \
"""\
    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];\
"""


            expected['outputs'] = \
"""\
    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_10
    + pydy_11 + pydy_4 + pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_10
    + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_11 + pydy_6 + pydy_7;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_11
    + pydy_7;\
"""

        self.generator._generate_cse()
        self.generator._generate_code_blocks()

        for k, v in self.generator.code_blocks.items():
            assert v == expected[k]

    def test_generate_code_blocks_without_cse(self):

        expected = {}

        expected['input_args'] = \
"""\
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],\
"""

        expected['output_args'] = \
"""\
              double output_0[36],
              double output_1[6]\
"""

        expected['input_docstring'] = \
"""\
input_0[19] : [c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, m0, m1, m2,
m3, m4, m5]
input_1[6] : [x0(t), x1(t), x2(t), x3(t), x4(t), x5(t)]
input_2[6] : [v0(t), v1(t), v2(t), v3(t), v4(t), v5(t)]
input_3[6] : [f0(t), f1(t), f2(t), f3(t), f4(t), f5(t)]\
"""

        expected['subexprs'] = \
"""    \
"""

        expected['outputs'] = \
"""\
    output_0[0] = input_0[13] + input_0[14] + input_0[15] + input_0[16] +
    input_0[17] + input_0[18];
    output_0[1] = input_0[14] + input_0[15] + input_0[16] + input_0[17] +
    input_0[18];
    output_0[2] = input_0[15] + input_0[16] + input_0[17] + input_0[18];
    output_0[3] = input_0[16] + input_0[17] + input_0[18];
    output_0[4] = input_0[17] + input_0[18];
    output_0[5] = input_0[18];
    output_0[6] = input_0[14] + input_0[15] + input_0[16] + input_0[17] +
    input_0[18];
    output_0[7] = input_0[14] + input_0[15] + input_0[16] + input_0[17] +
    input_0[18];
    output_0[8] = input_0[15] + input_0[16] + input_0[17] + input_0[18];
    output_0[9] = input_0[16] + input_0[17] + input_0[18];
    output_0[10] = input_0[17] + input_0[18];
    output_0[11] = input_0[18];
    output_0[12] = input_0[15] + input_0[16] + input_0[17] + input_0[18];
    output_0[13] = input_0[15] + input_0[16] + input_0[17] + input_0[18];
    output_0[14] = input_0[15] + input_0[16] + input_0[17] + input_0[18];
    output_0[15] = input_0[16] + input_0[17] + input_0[18];
    output_0[16] = input_0[17] + input_0[18];
    output_0[17] = input_0[18];
    output_0[18] = input_0[16] + input_0[17] + input_0[18];
    output_0[19] = input_0[16] + input_0[17] + input_0[18];
    output_0[20] = input_0[16] + input_0[17] + input_0[18];
    output_0[21] = input_0[16] + input_0[17] + input_0[18];
    output_0[22] = input_0[17] + input_0[18];
    output_0[23] = input_0[18];
    output_0[24] = input_0[17] + input_0[18];
    output_0[25] = input_0[17] + input_0[18];
    output_0[26] = input_0[17] + input_0[18];
    output_0[27] = input_0[17] + input_0[18];
    output_0[28] = input_0[17] + input_0[18];
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] +
    input_0[6] * input_0[14] + input_0[6] * input_0[15] + input_0[6] *
    input_0[16] + input_0[6] * input_0[17] + input_0[6] * input_0[18] -
    input_0[7] * input_1[0] + input_3[0] + input_3[1] + input_3[2] + input_3[3]
    + input_3[4] + input_3[5];
    output_1[1] = -input_0[1] * input_2[1] + input_0[6] * input_0[14] +
    input_0[6] * input_0[15] + input_0[6] * input_0[16] + input_0[6] *
    input_0[17] + input_0[6] * input_0[18] - input_0[8] * input_1[1] +
    input_3[1] + input_3[2] + input_3[3] + input_3[4] + input_3[5];
    output_1[2] = -input_0[2] * input_2[2] + input_0[6] * input_0[15] +
    input_0[6] * input_0[16] + input_0[6] * input_0[17] + input_0[6] *
    input_0[18] - input_0[9] * input_1[2] + input_3[2] + input_3[3] +
    input_3[4] + input_3[5];
    output_1[3] = -input_0[3] * input_2[3] + input_0[6] * input_0[16] +
    input_0[6] * input_0[17] + input_0[6] * input_0[18] - input_0[10] *
    input_1[3] + input_3[3] + input_3[4] + input_3[5];
    output_1[4] = -input_0[4] * input_2[4] + input_0[6] * input_0[17] +
    input_0[6] * input_0[18] - input_0[11] * input_1[4] + input_3[4] +
    input_3[5];
    output_1[5] = -input_0[5] * input_2[5] + input_0[6] * input_0[18] -
    input_0[12] * input_1[5] + input_3[5];\
"""

        self.generator._ignore_cse()
        self.generator._generate_code_blocks()

        for k, v in self.generator.code_blocks.items():
            assert v == expected[k]

    def test_doprint(self):

        expected_header = """\
void evaluate(
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],
              double output_0[36],
              double output_1[6]
             );
/*

input_0[19] : [c0, c1, c2, c3, c4, c5, g, k0, k1, k2, k3, k4, k5, m0, m1, m2,
m3, m4, m5]
input_1[6] : [x0(t), x1(t), x2(t), x3(t), x4(t), x5(t)]
input_2[6] : [v0(t), v1(t), v2(t), v3(t), v4(t), v5(t)]
input_3[6] : [f0(t), f1(t), f2(t), f3(t), f4(t), f5(t)]

*/\
"""
        if parse_version(SYMPY_VERSION) >= parse_version('1.2'):
            expected_source = """\
#include <math.h>
#include "boogly_bee.h"

void evaluate(
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],
              double output_0[36],
              double output_1[6]
             )
{

    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_0[6] * input_0[18] + input_3[5];
    double pydy_5 = input_0[6] * input_0[17] + pydy_4 + input_3[4];
    double pydy_6 = input_0[6] * input_0[16] + pydy_5 + input_3[3];
    double pydy_7 = input_0[6] * input_0[15] + pydy_6 + input_3[2];
    double pydy_8 = input_0[6] * input_0[14] + pydy_7 + input_3[1];

    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_8 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_8;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_7;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_6;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_5;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_4;

}\
"""
        elif parse_version(SYMPY_VERSION) >= parse_version('1.1'):
            expected_source = """\
#include <math.h>
#include "boogly_bee.h"

void evaluate(
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],
              double output_0[36],
              double output_1[6]
             )
{

    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];

    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_10
    + pydy_11 + pydy_4 + pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_10
    + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_11 + pydy_6 + pydy_7;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_11
    + pydy_7;

}\
"""
        elif parse_version(SYMPY_VERSION) > parse_version('1.0'):
            expected_source = """\
#include <math.h>
#include "boogly_bee.h"

void evaluate(
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],
              double output_0[36],
              double output_1[6]
             )
{

    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];
    double pydy_13 = pydy_10 + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    double pydy_14 = pydy_11 + pydy_7;

    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_13
    + pydy_4 + pydy_8;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] +
    pydy_13;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_14 + pydy_6;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] +
    pydy_14;

}\
"""
        else:
            expected_source = """\
#include <math.h>
#include "boogly_bee.h"

void evaluate(
              double input_0[19],
              double input_1[6],
              double input_2[6],
              double input_3[6],
              double output_0[36],
              double output_1[6]
             )
{

    double pydy_0 = input_0[17] + input_0[18];
    double pydy_1 = input_0[16] + pydy_0;
    double pydy_2 = input_0[15] + pydy_1;
    double pydy_3 = input_0[14] + pydy_2;
    double pydy_4 = input_3[2];
    double pydy_5 = input_3[3];
    double pydy_6 = input_3[4];
    double pydy_7 = input_3[5];
    double pydy_8 = input_0[6] * input_0[15];
    double pydy_9 = input_0[6] * input_0[16];
    double pydy_10 = input_0[6] * input_0[17];
    double pydy_11 = input_0[6] * input_0[18];
    double pydy_12 = input_0[6] * input_0[14] + pydy_10 + pydy_11 + pydy_4 +
    pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9 + input_3[1];

    output_0[0] = input_0[13] + pydy_3;
    output_0[1] = pydy_3;
    output_0[2] = pydy_2;
    output_0[3] = pydy_1;
    output_0[4] = pydy_0;
    output_0[5] = input_0[18];
    output_0[6] = pydy_3;
    output_0[7] = pydy_3;
    output_0[8] = pydy_2;
    output_0[9] = pydy_1;
    output_0[10] = pydy_0;
    output_0[11] = input_0[18];
    output_0[12] = pydy_2;
    output_0[13] = pydy_2;
    output_0[14] = pydy_2;
    output_0[15] = pydy_1;
    output_0[16] = pydy_0;
    output_0[17] = input_0[18];
    output_0[18] = pydy_1;
    output_0[19] = pydy_1;
    output_0[20] = pydy_1;
    output_0[21] = pydy_1;
    output_0[22] = pydy_0;
    output_0[23] = input_0[18];
    output_0[24] = pydy_0;
    output_0[25] = pydy_0;
    output_0[26] = pydy_0;
    output_0[27] = pydy_0;
    output_0[28] = pydy_0;
    output_0[29] = input_0[18];
    output_0[30] = input_0[18];
    output_0[31] = input_0[18];
    output_0[32] = input_0[18];
    output_0[33] = input_0[18];
    output_0[34] = input_0[18];
    output_0[35] = input_0[18];

    output_1[0] = -input_0[0] * input_2[0] + input_0[6] * input_0[13] -
    input_0[7] * input_1[0] + pydy_12 + input_3[0];
    output_1[1] = -input_0[1] * input_2[1] - input_0[8] * input_1[1] + pydy_12;
    output_1[2] = -input_0[2] * input_2[2] - input_0[9] * input_1[2] + pydy_10
    + pydy_11 + pydy_4 + pydy_5 + pydy_6 + pydy_7 + pydy_8 + pydy_9;
    output_1[3] = -input_0[3] * input_2[3] - input_0[10] * input_1[3] + pydy_10
    + pydy_11 + pydy_5 + pydy_6 + pydy_7 + pydy_9;
    output_1[4] = -input_0[4] * input_2[4] - input_0[11] * input_1[4] + pydy_10
    + pydy_11 + pydy_6 + pydy_7;
    output_1[5] = -input_0[5] * input_2[5] - input_0[12] * input_1[5] + pydy_11
    + pydy_7;

}\
"""

        header, source = self.generator.doprint()

        assert header == expected_header
        lines = expected_source.split('\n')
        assert source == '\n'.join(lines[:1] + lines[2:])

        header, source = self.generator.doprint(prefix=self.prefix)

        assert header == expected_header
        assert source == expected_source

    def test_write(self):

        header, source = self.generator.doprint(prefix=self.prefix)

        self.generator.write(self.prefix)

        with open(self.prefix + '.h') as f:
            assert f.read() == header

        with open(self.prefix + '.c') as f:
            assert f.read() == source

    def teardown(self):

        if os.path.isfile(self.prefix + '.h'):
            os.remove(self.prefix + '.h')

        if os.path.isfile(self.prefix + '.c'):
            os.remove(self.prefix + '.c')



#!/usr/bin/env python

from random import choice
import warnings

import numpy as np
import scipy as sp
import sympy as sm
from pydy.codegen.ode_function_generators import generate_ode_function

Cython = sm.external.import_module('Cython')
theano = sm.external.import_module('theano')

from ... import models
from ..ode_function_generators import (ODEFunctionGenerator,
                                       LambdifyODEFunctionGenerator,
                                       CythonODEFunctionGenerator,
                                       TheanoODEFunctionGenerator)

from ...utils import PyDyImportWarning

warnings.simplefilter('once', PyDyImportWarning)


def test_symbolic_lusolve_full_mass_matrix():
    sys = models.n_link_pendulum_on_cart(n=5, cart_force=False,
                                         joint_torques=False)

    g_symbolic_solve = CythonODEFunctionGenerator(
        sys.eom_method.forcing_full,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix_full,
        linear_sys_solver='sympy')
    rhs_symbolic_solve = g_symbolic_solve.generate()

    g_numeric_solve = CythonODEFunctionGenerator(
        sys.eom_method.forcing_full,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix_full,
        linear_sys_solver='numpy')
    rhs_numeric_solve = g_numeric_solve.generate()

    x = np.random.random(g_symbolic_solve.num_coordinates +
                         g_symbolic_solve.num_speeds)
    t = 5.125
    p = np.random.random(g_symbolic_solve.num_constants)

    np.testing.assert_allclose(rhs_numeric_solve(x, t, p),
                               rhs_symbolic_solve(x, t, p))


def test_symbolic_lusolve_min_mass_matrix():
    sys = models.n_link_pendulum_on_cart(n=5, cart_force=False,
                                         joint_torques=False)
    kin_diff_eqs = sys.eom_method.kindiffdict()
    coord_derivs = sm.Matrix([kin_diff_eqs[c.diff()] for c in
                              sys.coordinates])

    g_symbolic_solve = CythonODEFunctionGenerator(
        sys.eom_method.forcing,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix,
        coordinate_derivatives=coord_derivs,
        linear_sys_solver='sympy')
    rhs_symbolic_solve = g_symbolic_solve.generate()

    g_numeric_solve = CythonODEFunctionGenerator(
        sys.eom_method.forcing,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix,
        coordinate_derivatives=coord_derivs,
        linear_sys_solver='numpy')
    rhs_numeric_solve = g_numeric_solve.generate()

    x = np.random.random(g_symbolic_solve.num_coordinates +
                         g_symbolic_solve.num_speeds)
    t = 5.125
    p = np.random.random(g_symbolic_solve.num_constants)

    np.testing.assert_allclose(rhs_numeric_solve(x, t, p),
                               rhs_symbolic_solve(x, t, p))


def test_cse_same_numerical_results():
    # NOTE : This ensurses that the same results are always given for the sympy
    # cse outputs, which seem to change every version.

    if not Cython:
        return

    sys = models.n_link_pendulum_on_cart(n=5, cart_force=False,
                                         joint_torques=False)

    g_no_cse = CythonODEFunctionGenerator(
        sys.eom_method.forcing_full,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix_full,
        cse=False)
    rhs_func_no_cse = g_no_cse.generate()

    g_cse = CythonODEFunctionGenerator(
        sys.eom_method.forcing_full,
        sys.coordinates,
        sys.speeds,
        sys.constants_symbols,
        mass_matrix=sys.eom_method.mass_matrix_full,
        cse=True)
    rhs_func_cse = g_cse.generate()

    x = np.random.random(g_cse.num_coordinates + g_cse.num_speeds)
    t = 5.125
    p = np.random.random(g_cse.num_constants)

    np.testing.assert_allclose(rhs_func_no_cse(x, t, p),
                               rhs_func_cse(x, t, p))


class TestODEFunctionGenerator(object):

    def setup(self):

        self.sys = models.multi_mass_spring_damper(2)
        self.rhs = self.sys.eom_method.rhs()
        self.generator = ODEFunctionGenerator(self.rhs,
                                              self.sys.coordinates,
                                              self.sys.speeds,
                                              self.sys.constants_symbols)

    def test_init_full_rhs(self):

        g = ODEFunctionGenerator(self.rhs,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols)

        assert g.num_coordinates == 2
        assert g.num_speeds == 2
        assert g.num_states == 4
        assert g.system_type == 'full rhs'

    def test_init_full_mass_matrix(self):

        g = ODEFunctionGenerator(self.sys.eom_method.forcing_full,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix_full)

        assert g.num_coordinates == 2
        assert g.num_speeds == 2
        assert g.num_states == 4
        assert g.system_type == 'full mass matrix'

    def test_init_min_mass_matrix(self):

        kin_diff_eqs = self.sys.eom_method.kindiffdict()
        coord_derivs = sm.Matrix([kin_diff_eqs[c.diff()] for c in
                                  self.sys.coordinates])

        g = ODEFunctionGenerator(self.sys.eom_method.forcing,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix,
                                 coordinate_derivatives=coord_derivs)

        assert g.num_coordinates == 2
        assert g.num_speeds == 2
        assert g.num_states == 4
        assert g.system_type == 'min mass matrix'

    def test_set_linear_system_solver(self):

        g = ODEFunctionGenerator(self.sys.eom_method.forcing_full,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix_full)

        assert g.linear_sys_solver == 'numpy'
        assert g._solve_linear_system == np.linalg.solve

        g = ODEFunctionGenerator(self.sys.eom_method.forcing_full,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix_full,
                                 linear_sys_solver='numpy')

        assert g.linear_sys_solver == 'numpy'
        assert g._solve_linear_system == np.linalg.solve

        g = ODEFunctionGenerator(self.sys.eom_method.forcing_full,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix_full,
                                 linear_sys_solver='scipy')

        assert g.linear_sys_solver == 'scipy'
        assert g._solve_linear_system == sp.linalg.solve

        solver = lambda A, b: np.dot(np.inv(A), b)

        g = ODEFunctionGenerator(self.sys.eom_method.forcing_full,
                                 self.sys.coordinates,
                                 self.sys.speeds,
                                 self.sys.constants_symbols,
                                 mass_matrix=self.sys.eom_method.mass_matrix_full,
                                 linear_sys_solver=solver)

        assert g.linear_sys_solver == solver
        assert g._solve_linear_system == solver

    def test_no_constants(self):
        sys = models.multi_mass_spring_damper()
        constant_vals = {sm.Symbol('m0'): 1.0, sm.Symbol('c0'): 2.0, sm.Symbol('k0'): 3.0}

        sym_rhs = sys.eom_method.rhs()

        rhs = generate_ode_function(sym_rhs, sys.coordinates, sys.speeds, constant_vals)
        rhs2 = generate_ode_function(sym_rhs.subs(constant_vals), sys.coordinates, sys.speeds)

        assert np.array_equal(rhs(np.array([1.0, 2.0]), 0.0, constant_vals),
                              rhs2(np.array([1.0, 2.0]), 0.0))


class TestODEFunctionGeneratorSubclasses(object):

    ode_function_subclasses = [LambdifyODEFunctionGenerator]

    if Cython:
        ode_function_subclasses.append(CythonODEFunctionGenerator)
    else:
        warnings.warn("Cython was not found so the related tests are being"
                      " skipped.", PyDyImportWarning)

    if theano:
        ode_function_subclasses.append(TheanoODEFunctionGenerator)
    else:
        warnings.warn("Theano was not found so the related tests are being"
                      " skipped.", PyDyImportWarning)

    def setup(self):

        self.sys = models.multi_mass_spring_damper()
        # Best keep these in order, otherwise it may change between SymPy
        # versions.
        self.constants = list(sm.ordered(self.sys.constants_symbols))

    def eval_rhs(self, rhs):

        # In order:
        c0 = 1.0
        k0 = 2.0
        m0 = 3.0

        x0 = 1.0
        v0 = 2.0

        xdot = rhs(np.array([x0, v0]), 0.0, np.array([c0, k0, m0]))

        expected_xdot = np.array([v0, (-c0 * v0 - k0 * x0) / m0])

        np.testing.assert_allclose(xdot, expected_xdot)

    def test_init_doc(self):

        for Subclass in self.ode_function_subclasses:
            assert (Subclass.__init__.__doc__ ==
                    ODEFunctionGenerator.__init__.__doc__)

    def test_generate_full_rhs(self):

        rhs = self.sys.eom_method.rhs()

        for Subclass in self.ode_function_subclasses:

            g = Subclass(rhs,
                         self.sys.coordinates,
                         self.sys.speeds,
                         self.constants)

            rhs_func = g.generate()

            self.eval_rhs(rhs_func)

    def test_generate_full_mass_matrix(self):

        for Subclass in self.ode_function_subclasses:

            g = Subclass(self.sys.eom_method.forcing_full,
                         self.sys.coordinates,
                         self.sys.speeds,
                         self.constants,
                         mass_matrix=self.sys.eom_method.mass_matrix_full)

            rhs_func = g.generate()

            self.eval_rhs(rhs_func)

    def test_generate_min_mass_matrix(self):

        kin_diff_eqs = self.sys.eom_method.kindiffdict()
        coord_derivs = sm.Matrix([kin_diff_eqs[c.diff()] for c in
                                  self.sys.coordinates])

        for Subclass in self.ode_function_subclasses:

            g = Subclass(self.sys.eom_method.forcing,
                         self.sys.coordinates,
                         self.sys.speeds,
                         self.constants,
                         mass_matrix=self.sys.eom_method.mass_matrix,
                         coordinate_derivatives=coord_derivs)

            rhs_func = g.generate()

            self.eval_rhs(rhs_func)

    def test_generate_with_specifieds(self):

        # This model has three specifieds.
        sys = models.n_link_pendulum_on_cart(2, True, True)

        kin_diff_eqs = sys.eom_method.kindiffdict()
        coord_derivs = sm.Matrix([kin_diff_eqs[c.diff()] for c in
                                  sys.coordinates])

        rhs = sys.eom_method.rhs()

        common_args = [sys.coordinates, sys.speeds, sys.constants_symbols]
        args_dct = {}
        args_dct['full rhs'] = [rhs] + common_args
        args_dct['full mass matrix'] = [sys.eom_method.forcing_full] + common_args
        args_dct['min mass matrix'] = [sys.eom_method.forcing] + common_args

        kwargs_dct = {}
        kwargs_dct['full rhs'] = {}
        kwargs_dct['full mass matrix'] = \
            {'mass_matrix': sys.eom_method.mass_matrix_full}
        kwargs_dct['min mass matrix'] = \
            {'mass_matrix': sys.eom_method.mass_matrix,
             'coordinate_derivatives': coord_derivs}

        for Subclass in self.ode_function_subclasses:
            for system_type, args in args_dct.items():

                g = Subclass(*args, specifieds=sys.specifieds_symbols,
                             **kwargs_dct[system_type])

                f = g.generate()

                rand = np.random.random

                x = rand(g.num_states)
                r = rand(g.num_specifieds)
                p = rand(g.num_constants)

                subs = {}
                for arr, syms in zip([x, r, p], [sys.states,
                                                 sys.specifieds_symbols,
                                                 sys.constants_symbols]):
                    for val, sym in zip(arr, syms):
                        subs[sym] = val

                try:
                    expected = sm.matrix2numpy(rhs.subs(subs),
                                               dtype=float).squeeze()
                except TypeError:
                    # Earlier SymPy versions don't support the dtype kwarg.
                    expected = np.asarray(sm.matrix2numpy(rhs.subs(subs)),
                                          dtype=float).squeeze()

                xdot = f(x, 0.0, r, p)

                np.testing.assert_allclose(xdot, expected)

    def test_rhs_args(self):
        # This test takes a while to run but it checks all the combinations.

        # There are eight constants and four specified inputs available.
        sys = models.n_link_pendulum_on_cart(3, True, True)
        right_hand_side = sys.eom_method.rhs()
        constants = list(sm.ordered(sys.constants_symbols))
        specifieds = list(sm.ordered(sys.specifieds_symbols))

        constants_arg_types = [None, 'array', 'dictionary']
        specifieds_arg_types = [None, 'array', 'function', 'dictionary']

        p_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
        p_dct = dict(zip(constants, p_array))

        p = {}
        p[None] = choice([p_array, p_dct])
        p['array'] = p_array
        p['dictionary'] = p_dct

        r_array = np.array([1.0, 2.0, 3.0, 4.0])
        r_dct_1 = dict(zip(specifieds, r_array))
        r_dct_2 = {tuple(specifieds):
                   lambda x, t: r_array}
        r_dct_3 = {specifieds[0]: lambda x, t: np.ones(1),
                   (specifieds[3], specifieds[1]):
                   lambda x, t: np.array([4.0, 2.0]),
                   specifieds[2]: 3.0 * np.ones(1)}
        r_func = lambda x, t: np.array([1.0, 2.0, 3.0, 4.0])

        r = {}
        r[None] = choice([r_array, r_dct_1, r_dct_2, r_dct_3, r_func])
        r['array'] = r_array
        r['dictionary'] = choice([r_dct_1, r_dct_2, r_dct_3])
        r['function'] = r_func

        x = np.random.random(len(sys.states))

        for p_arg_type in constants_arg_types:
            for r_arg_type in specifieds_arg_types:

                g = LambdifyODEFunctionGenerator(right_hand_side,
                                                 sys.coordinates,
                                                 sys.speeds,
                                                 constants,
                                                 specifieds=specifieds,
                                                 constants_arg_type=p_arg_type,
                                                 specifieds_arg_type=r_arg_type)
                rhs = g.generate()

                xdot = rhs(x, 0.0, r[r_arg_type], p[p_arg_type])

                try:
                    np.testing.assert_allclose(xdot, last_xdot)
                except NameError:
                    pass

                last_xdot = xdot

        # Now make sure it all works with specifieds=None
        sys = models.n_link_pendulum_on_cart(3, False, False)
        right_hand_side = sys.eom_method.rhs()
        constants = list(sm.ordered(sys.constants_symbols))

        del last_xdot

        for p_arg_type in constants_arg_types:
            for r_arg_type in specifieds_arg_types:

                g = LambdifyODEFunctionGenerator(right_hand_side,
                                                 sys.coordinates,
                                                 sys.speeds,
                                                 constants,
                                                 constants_arg_type=p_arg_type,
                                                 specifieds_arg_type=r_arg_type)

                assert g.specifieds_arg_type is None

                rhs = g.generate()

                xdot = rhs(x, 0.0, p[p_arg_type])

                try:
                    np.testing.assert_allclose(xdot, last_xdot)
                except NameError:
                    pass

                last_xdot = xdot

    def test_rhs_docstring(self):

        sys = models.n_link_pendulum_on_cart(2, False, False)
        right_hand_side = sys.eom_method.rhs()

        constants = list(sm.ordered(sys.constants_symbols))

        constants_arg_types = [None, 'array', 'dictionary']

        rhs_doc_template = \
"""\
Returns the derivatives of the states, i.e. numerically evaluates the right
hand side of the first order differential equation.

x' = f(x, t,{specified_call_sig} p)

Parameters
==========
x : ndarray, shape(6,)
    The state vector is ordered as such:
        - q0(t)
        - q1(t)
        - q2(t)
        - u0(t)
        - u1(t)
        - u2(t)
t : float
    The current time.{specifieds_explanation}{constants_explanation}
Returns
=======
dx : ndarray, shape(6,)
    The derivative of the state vector.

"""

        constants_doc_templates = {}

        constants_doc_templates['dictionary'] = \
"""
p : dictionary len(6)
    A dictionary that maps the constants symbols to their numerical values
    with at least these keys:
        - g
        - l0
        - l1
        - m0
        - m1
        - m2
"""

        constants_doc_templates['array'] = \
"""
p : ndarray shape(6,)
    A ndarray of floats that give the numerical values of the constants in
    this order:
        - g
        - l0
        - l1
        - m0
        - m1
        - m2
"""

        constants_doc_templates[None] = \
"""
p : dictionary len(6) or ndarray shape(6,)
    Either a dictionary that maps the constants symbols to their numerical
    values or an array with the constants in the following order:
        - g
        - l0
        - l1
        - m0
        - m1
        - m2
"""

        for p_arg_type in constants_arg_types:

            _rhs_doc_template = rhs_doc_template.format(**{
                'specified_call_sig': '',
                'specifieds_explanation': '',
                'constants_explanation': constants_doc_templates[p_arg_type]
                })

            g = LambdifyODEFunctionGenerator(right_hand_side,
                                             sys.coordinates,
                                             sys.speeds,
                                             constants,
                                             constants_arg_type=p_arg_type)

            rhs = g.generate()

            assert (_rhs_doc_template == rhs.__doc__)

        sys = models.n_link_pendulum_on_cart(2, True, True)
        right_hand_side = sys.eom_method.rhs()

        constants = list(sm.ordered(sys.constants_symbols))
        specifieds = list(sm.ordered(sys.specifieds_symbols))

        specifieds_arg_types = [None, 'array', 'function', 'dictionary']

        specifieds_doc_templates = {}

        specifieds_doc_templates[None] = \
"""
r : dictionary; ndarray, shape(3,); function

    There are three options for this argument. (1) is more flexible but
    (2) and (3) are much more efficient.

    (1) A dictionary that maps the specified functions of time to floats,
    ndarrays, or functions that produce ndarrays. The keys can be a single
    specified symbolic function of time or a tuple of symbols. The total
    number of symbols must be equal to 3. If the value is a
    function it must be of the form g(x, t), where x is the current state
    vector ndarray and t is the current time float and it must return an
    ndarray of the correct shape. For example::

      r = {a: 1.0,
           (d, b) : np.array([1.0, 2.0]),
           (e, f) : lambda x, t: np.array(x[0], x[1]),
           c: lambda x, t: np.array(x[2])}

    (2) A ndarray with the specified values in the correct order and of the
    correct shape.

    (3) A function that must be of the form g(x, t), where x is the current
    state vector and t is the current time and it must return an ndarray of
    the correct shape.

    The specified inputs are, in order:
        - F(t)
        - T1(t)
        - T2(t)"""

        specifieds_doc_templates['array'] = \
"""
r : ndarray, shape(3,)

    A ndarray with the specified values in the correct order and of the
    correct shape.

    The specified inputs are, in order:
        - F(t)
        - T1(t)
        - T2(t)"""

        specifieds_doc_templates['dictionary'] = \
"""
r : dictionary

    A dictionary that maps the specified functions of time to floats,
    ndarrays, or functions that produce ndarrays. The keys can be a single
    specified symbolic function of time or a tuple of symbols. The total
    number of symbols must be equal to 3. If the value is a
    function it must be of the form g(x, t), where x is the current state
    vector ndarray and t is the current time float and it must return an
    ndarray of the correct shape. For example::

      r = {a: 1.0,
           (d, b) : np.array([1.0, 2.0]),
           (e, f) : lambda x, t: np.array(x[0], x[1]),
           c: lambda x, t: np.array(x[2])}

    The specified inputs are, in order:
        - F(t)
        - T1(t)
        - T2(t)"""

        specifieds_doc_templates['function'] = \
"""
r : function

    A function that must be of the form g(x, t), where x is the current
    state vector and t is the current time and it must return an ndarray of
    shape(3,).

    The specified inputs are, in order:
        - F(t)
        - T1(t)
        - T2(t)"""

        for p_arg_type in constants_arg_types:
            for r_arg_type in specifieds_arg_types:

                _rhs_doc_template = rhs_doc_template.format(**{
                    'specified_call_sig': ' r,',
                    'specifieds_explanation': specifieds_doc_templates[r_arg_type],
                    'constants_explanation': constants_doc_templates[p_arg_type]
                    })

                g = LambdifyODEFunctionGenerator(right_hand_side,
                                                 sys.coordinates,
                                                 sys.speeds,
                                                 constants,
                                                 specifieds=specifieds,
                                                 constants_arg_type=p_arg_type,
                                                 specifieds_arg_type=r_arg_type)

                rhs = g.generate()

                assert (_rhs_doc_template == rhs.__doc__)



#!/usr/bin/env python

# standard library
from __future__ import division
import os
import sys
import warnings
import json
import distutils
import distutils.dir_util
import datetime
from collections import OrderedDict

# external
from pkg_resources import parse_version
import numpy as np
from sympy import latex
from sympy.physics.mechanics import ReferenceFrame, Point, dynamicsymbols
try:
    import pythreejs as p3js
except ImportError:
    p3js = None

# local
from .camera import PerspectiveCamera
from .server import Server
from .light import PointLight
from ..system import System
from ..utils import PyDyImportWarning

raw_input = input

__all__ = ['Scene']

warnings.simplefilter('once', PyDyImportWarning)

try:
    import IPython
    if parse_version(IPython.__version__) < parse_version('3.0'):
        msg = ('PyDy only supports IPython >= 3.0.0. You have '
               'IPython {} installed. IPython related functionalities will '
               'not be available')
        warnings.warn(msg.format(IPython.__version__), PyDyImportWarning)
        ipython_less_than_3 = True
    else:
        ipython_less_than_3 = False
        # If IPython >= 4.0 use ipywidgets if installed to avoid the
        # deprecation warning.
        try:
            import ipywidgets as widgets
        except ImportError:
            from IPython.html import widgets
        from IPython.display import display, Javascript
except ImportError:
    IPython = None


class Scene(object):
    """The Scene class generates all of the data required for animating a
    set of visualization frames.

    """

    pydy_directory = "pydy-resources"

    def __init__(self, reference_frame, origin, *visualization_frames,
                 **kwargs):
        """Initialize a Scene instance.

        Parameters
        ==========
        reference_frame : sympy.physics.mechanics.ReferenceFrame
            The base reference frame for the scene. The motion of all of the
            visualization frames, cameras, and lights will be generated with
            respect to this reference frame.
        origin : sympy.physics.mechanics.Point
            The base point for the scene. The motion of all of the
            visualization frames, cameras, and lights will be generated with
            respect to this point.
        visualization_frames : VisualizationFrame
            One or more visualization frames which are to be displayed in
            the scene.
        name : string, optional, default='unnamed'
            Name of Scene object.
        cameras : list of Camera instances, optional
            The cameras with which to display the object. The first camera
            is used to display the scene initially. The default is a single
            PerspectiveCamera tied to the base reference frame and
            positioned away from the origin along the reference frame's z
            axis.
        lights : list of Light instances, optional
            The lights used in the scene. The default is a single Light tied
            to the base reference frame and positioned away from the origin
            along the reference frame's z axis at the same point as the
            default camera.
        system : System, optional, default=None
            A PyDy system class which is initiated such that the
            ``integrate()`` method will produce valid state trajectories.
        times : array_like, shape(n,), optional, default=None
            Monotoncially increaing float values of time that correspond to
            the state trajectories.
        constants : dictionary, optional, default=None
            A dictionary that maps SymPy symbols to floats. This should
            contain at least all necessary symbols to evaluate the
            transformation matrices of the visualization frame, cameras, and
            lights and to evaluate the Shapes' parameters.
        states_symbols : sequence of functions, len(m), optional, default=None
            An ordered sequence of the SymPy functions that represent the
            states. The order must match the order of the
            ``states_trajectories``.
        states_trajectories : array_like, shape(n, m), optional, default=None
            A two dimensional array with numerical values for each state at
            each point in time during the animation.

        Notes
        =====
        The user is allowed to supply either system or times, constants,
        states_symbols, and states_trajectories. Providing a System allows for
        interactively changing the simulation parameters via the Scene GUI
        in the IPython notebook.

        """

        self.reference_frame = reference_frame
        self.origin = origin
        self.visualization_frames = list(visualization_frames)

        vec = 10 * self.reference_frame.z
        self._default_camera_point = self.origin.locatenew('p_camera', vec)
        self._default_light_point = self.origin.locatenew('p_light', vec)

        default_kwargs = {'name': 'unnamed',
                          'cameras': [PerspectiveCamera('DefaultCamera',
                                                        self.reference_frame,
                                                        self._default_camera_point)],
                          'lights': [PointLight('DefaultLight',
                                                self.reference_frame,
                                                self._default_light_point)],
                          'system': None,
                          'times': None,
                          'constants': None,
                          'states_symbols': None,
                          'states_trajectories': None,
                          'frames_per_second': 30}
        default_kwargs.update(kwargs)

        for k, v in default_kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        """Returns the name of the scene."""
        return self._name

    @name.setter
    def name(self, new_name):
        """Sets the name of the scene."""
        if not isinstance(new_name, str):
            raise TypeError("'name' should be a valid string.")
        else:
            self._name = new_name

    @property
    def origin(self):
        """Returns the origin point of the scene."""
        return self._origin

    @origin.setter
    def origin(self, new_origin):
        """Sets the origin point of the scene."""
        if not isinstance(new_origin, Point):
            raise TypeError("'origin' should be a valid Point object.")
        else:
            self._origin = new_origin

    @property
    def reference_frame(self):
        """Returns the base reference frame of the scene."""
        return self._reference_frame

    @reference_frame.setter
    def reference_frame(self, new_reference_frame):
        """Sets the base reference frame for the scene."""
        if not isinstance(new_reference_frame, ReferenceFrame):
            raise TypeError("'reference_frame' should be a valid "
                            "ReferenceFrame object.")
        else:
            self._reference_frame = new_reference_frame

    @property
    def system(self):
        return self._system

    @system.setter
    def system(self, new_system):

        if new_system is not None and not isinstance(new_system, System):
            msg = "{} should be a valid pydy.System object".format(new_system)
            raise TypeError(msg)

        if new_system is not None:
            msg = ('The {} attribute has already been set, so the system '
                   'cannot be set. Use the clear_trajectories method to '
                   'set all relevant attributes to None.')
            for attr in ['times', 'constants', 'states_symbols',
                         'states_trajectories']:
                try:
                    if getattr(self, attr) is not None:
                        raise ValueError(msg.format(attr))
                except AttributeError:
                    pass

        self._system = new_system

    @property
    def times(self):
        if self.system is not None:
            return self.system.times
        else:
            return self._times

    @times.setter
    def times(self, new_times):

        try:
            if new_times is not None and self.system is not None:
                msg = ('The system attribute has already been set, so the '
                       'times cannot be set. Set Scene.system = None to '
                       'allow a time array to be added.')
                raise ValueError(msg)
        except AttributeError:
            pass

        try:
            if new_times is not None and self.states_trajectories is not None:
                len_traj = self.states_trajectories.shape[0]
                if len(new_times) != len_traj:
                    msg = ('The times array length, {}, does not match the '
                           'length of the state trajectories array, {}.')
                    raise ValueError(msg.format(len(new_times), len_traj))
        except AttributeError:
            pass

        if new_times is None:
            self._times = new_times
        else:
            self._times = np.array(new_times)

    @property
    def states_symbols(self):
        if self.system is not None:
            return self.system.states
        else:
            return self._states_symbols

    @states_symbols.setter
    def states_symbols(self, new_states_symbols):

        try:
            if new_states_symbols is not None and self.system is not None:
                msg = ('The system attribute has already been set, so the '
                       'coordinates cannot be set. Set Scene.system = None '
                       'to allow a coordinates array to be added.')
                raise ValueError(msg)
        except AttributeError:
            pass

        try:
            if (new_states_symbols is not None and self.states_trajectories
                is not None):
                len_traj = self.states_trajectories.shape[1]
                if len(new_states_symbols) != len_traj:
                    msg = ('The number of states, {}, does not match the '
                           'number of states present in the state '
                           'trajectories array, {}.')
                    raise ValueError(msg.format(len(new_states_symbols), len_traj))
        except AttributeError:
            pass

        self._states_symbols = new_states_symbols

    @property
    def states_trajectories(self):
        if self.system is not None:
            # TODO : This calculation needs to be cached in some way.
            return self.system.integrate()
        else:
            return self._states_trajectories

    @states_trajectories.setter
    def states_trajectories(self, new_states_trajectories):

        try:
            if new_states_trajectories is not None and self.system is not None:
                msg = ('The system attribute has already been set, so the '
                       'states_trajectories cannot be set. Set Scene.system '
                       '= None to allow a states_trajectories array to be '
                       'added.')
                raise ValueError(msg)
        except AttributeError:
            pass

        try:
            if new_states_trajectories is not None and self.times is not None:
                if len(self.times) != new_states_trajectories.shape[0]:
                    msg = ("The number of time instances do not match the "
                           "number in the times array.")
                    raise ValueError(msg)
        except AttributeError:
            pass

        try:
            if (new_states_trajectories is not None and self.states_symbols
                is not None):
                if new_states_trajectories.shape[1] != len(self.states_symbols):
                    msg = ("The number of states in the trajectory do not "
                           "match the number of states symbols.")
                    raise ValueError(msg)
        except AttributeError:
            pass

        self._states_trajectories = new_states_trajectories

    @property
    def constants(self):
        if self.system is not None:
            return self.system.constants
        else:
            return self._constants

    @constants.setter
    def constants(self, new_constants):
        try:
            if new_constants is not None and self.system is not None:
                msg = ('The system attribute has already been set, so the '
                       'constants cannot be set. Set Scene.system = None to '
                       'allow a constants array to be added.')
                raise ValueError(msg)
        except AttributeError:
            pass
        self._constants = new_constants

    def clear_trajectories(self):
        """Sets the 'system', 'times', 'constants', 'states_symbols', and
        'states_trajectories' to None."""
        for attr in ['system', 'times', 'constants', 'states_symbols',
                     'states_trajectories']:
            setattr(self, attr, None)

    def _generate_json(self, directory=None, prefix=None):
        """Creates two JSON files and copies all the necessary static files
        in the specified directory. One of the JSON files contains the scene
        information and other one contains the simulation data.

        Parameters
        ==========
        directory : string, optional, default=None
            The directory in which the json files are placed. If None, this
            will be the current working directory.
        prefix : string
            A custom prefix for the two json files. If None, a time stamp is
            used.

        """

        if directory is None:
            directory = os.getcwd()

        if prefix is None:
            prefix = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

        self._scene_json_file = prefix + "_scene_desc.json"
        self._simulation_json_file = prefix + "_simulation_data.json"

        if self.system is None:
            constants = self.constants
            num_time_steps = self.states_trajectories.shape[0]
        else:
            constants = self.system.constants
            num_time_steps = len(self.system.times)
        # TODO : This assumes that all constants have unique strings and
        # that they are valid strings for the JSON file which can possibly
        # fail.
        constant_map_for_json = {str(k): v for k, v in constants.items()}

        self._generate_simulation_dict()
        self._generate_scene_dict()

        self._scene_info["simulationData"] = self._simulation_json_file

        times = None
        if self.times is not None:
            times = self.times
        elif self.system and self.system.times is not None:
            times = self.system.times

        if times is None:
            self._scene_info["timeDelta"] = 1.0 / self.frames_per_second
            self._scene_info["startTime"] = 0.0
        else:
            # Assume that times is evenly spaced and monotonic.
            # TODO: Interpolate if times are not evenly spaced.
            total_time = times[-1] - times[0]
            self._scene_info["timeDelta"] = total_time / (num_time_steps - 1)
            self._scene_info["startTime"] = times[0]

        self._scene_info["fps"] = self.frames_per_second
        self._scene_info["speedup"] = (self.frames_per_second *
                                       self._scene_info["timeDelta"])
        self._scene_info["timeSteps"] = num_time_steps
        self._scene_info["constant_map"] = constant_map_for_json

        scene_file_path = os.path.join(directory, self._scene_json_file)
        with open(scene_file_path, 'w') as scene_data_outfile:
            scene_data_outfile.write(json.dumps(self._scene_info,
                                                indent=4,
                                                separators=(',', ': ')))

        sim_file_path = os.path.join(directory, self._simulation_json_file)
        with open(sim_file_path, 'w') as simulation_data_outfile:
            simulation_data_outfile.write(json.dumps(
                self._simulation_info, indent=4,
                separators=(',', ': ')))

    def _generate_scene_dict(self):
        """Generates a dictionary containing all of the information needed
        to build the scene."""

        self._scene_info = {}
        self._scene_info["source"] = "PyDy"
        self._scene_info["name"] = self.name
        self._scene_info["newtonian_frame"] = str(self.reference_frame)
        # TODO : This should be accomodated in scene instead of width/height
        # of scene.
        self._scene_info["workspaceSize"] = 0.2

        self._scene_info["objects"] = {}
        self._scene_info["cameras"] = {}
        self._scene_info["lights"] = {}

        for frame in self.visualization_frames:
            constants = self.constants
            object_info = frame.generate_scene_dict(constant_map=constants)
            self._scene_info["objects"].update(object_info)

        for camera in self.cameras:
            object_info = camera.generate_scene_dict()
            self._scene_info["cameras"].update(object_info)

        for light in self.lights:
            object_info = light.generate_scene_dict()
            self._scene_info["lights"].update(object_info)

    def _generate_meshes_tracks(self):
        """Creates KeyFrameTrack for each visualization frame."""

        self._meshes = []
        self._tracks = []

        traj = self.states_trajectories

        for vizframe in self.visualization_frames:
            vizframe.generate_transformation_matrix(self.reference_frame,
                                                    self.origin)
            vizframe.generate_numeric_transform_function(self.states_symbols,
                                                         self.constants.keys())
            vizframe._create_keyframetrack(self.times, traj,
                                           list(self.constants.values()),
                                           constant_map=self.constants)
            self._tracks.append(vizframe._track)
            self._meshes.append(vizframe._mesh)

    def display_jupyter(self, window_size=(800, 600), axes_arrow_length=None):
        """Returns a PyThreeJS Renderer and AnimationAction for displaying and
        animating the scene inside a Jupyter notebook.

        Parameters
        ==========
        window_size : 2-tuple of integers
            2-tuple containing the width and height of the renderer window in
            pixels.
        axes_arrow_length : float
            If a positive value is supplied a red (x), green (y), and blue (z)
            arrows of the supplied length will be displayed as arrows for the
            global axes.

        Returns
        =======
        vbox : widgets.VBox
            A vertical box containing the action (pythreejs.AnimationAction)
            and renderer (pythreejs.Renderer).

        """
        if p3js is None:
            raise ImportError('pythreejs needs to be installed.')

        self._generate_meshes_tracks()

        view_width = window_size[0]
        view_height = window_size[1]

        camera = p3js.PerspectiveCamera(position=[1, 1, 1],
                                        aspect=view_width/view_height)
        key_light = p3js.DirectionalLight()
        ambient_light = p3js.AmbientLight()

        children = self._meshes + [camera, key_light, ambient_light]

        if axes_arrow_length is not None:
            children += [p3js.AxesHelper(size=abs(axes_arrow_length))]

        scene = p3js.Scene(children=children)

        controller = p3js.OrbitControls(controlling=camera)
        renderer = p3js.Renderer(camera=camera, scene=scene,
                                 controls=[controller],
                                 width=view_width, height=view_height)

        clip = p3js.AnimationClip(tracks=self._tracks, duration=self.times[-1])
        action = p3js.AnimationAction(p3js.AnimationMixer(scene), clip, scene)

        return widgets.VBox([action, renderer])

    def _generate_simulation_dict(self):
        """Returns a dictionary containing all of the simulation
        information. It consists of all the simulation data along with
        references to the objects, for allowing motion to the objects in the
        PyDy visualizer.

        Notes
        =====

        This method must be called before ``generate_scene_dict``.

        """

        self._simulation_info = {}

        traj = self.states_trajectories

        for group in [self.visualization_frames, self.cameras, self.lights]:
            for frame in group:
                frame.generate_transformation_matrix(self.reference_frame,
                                                     self.origin)
                frame.generate_numeric_transform_function(self.states_symbols,
                                                          self.constants.keys())
                frame.evaluate_transformation_matrix(traj,
                                                     self.constants.values())
                self._simulation_info.update(frame.generate_simulation_dict())

    def generate_visualization_json_system(self, system, **kwargs):
        """Creates the visualization JSON files for the provided system.

        Parameters
        ==========
        system : pydy.system.System
            A fully developed PyDy system that is prepared for the
            ``.integrate()`` method.
        fps : int, optional, default=30
            Frames per second at which animation should be displayed. Please
            not that this should not exceed the hardware limit of the
            display device to be used. Default is 30fps.
        outfile_prefix : str, optional, default=None
            A prefix for the JSON files. The files will be named as
            `outfile_prefix_scene_desc.json` and
            `outfile_prefix_simulation_data.json`. If not specified a
            timestamp shall be used as the prefix.

        Notes
        =====
        The optional keyword arguments are the same as those in the
        ``generate_visualization_json`` method.

        """

        self.system = system
        prefix = None
        for k, v in kwargs.items():
            if k == 'fps':
                self.frames_per_second = v
            if k == 'outfile_prefix':
                prefix = v
        self._generate_json(prefix=prefix)

    def create_static_html(self, overwrite=False, silent=False, prefix=None):
        """Creates a directory named ``pydy-visualization`` in the current
        working directory which contains all of the HTML, CSS, Javascript,
        and json files required to run the vizualization application. To run
        the application, navigate into the ``pydy-visualization`` directory
        and start a webserver from that directory, e.g.::

            $ python -m SimpleHTTPServer

        Now, in a WebGL compliant browser, navigate to::

            http://127.0.0.1:8000

        to view and interact with the visualization.

        This method can also be used to output files for embedding the
        visualizations in static webpages. Simply copy the contents of
        static directory in the relevant directory for embedding in a static
        website.

        Parameters
        ----------
        overwrite : boolean, optional, default=False
            If True, the directory named ``pydy-visualization`` in the
            current working directory will be overwritten.
        silent : boolean, optional, default=False
            If True, no messages will be displayed to STDOUT.
        prefix : string, optional
            An optional prefix for the json data files.

        """

        pydy_dir = os.path.join(os.getcwd(), self.pydy_directory)

        if os.path.exists(pydy_dir) and overwrite is False:
            msg = ("The '{}' directory already exists. Would "
                   "you like to overwrite the contents? [y|n]\n")
            ans = raw_input(msg.format(self.pydy_directory))
            if ans == 'y':
                overwrite = True
            else:
                if not silent:
                    print("Aborted!")
                return

        # Copy all of the HTML/CSS/JS files from the source tree into the
        # local directory.
        if not silent:
            print("Copying static data.")
        src = os.path.join(os.path.dirname(__file__), 'static')
        distutils.dir_util.copy_tree(src, pydy_dir)

        # Add the two json files to the directory.
        if not silent:
            print("Copying Simulation data.")
        self._generate_json(directory=pydy_dir, prefix=prefix)

        if not silent:
            msg = ("To view the visualzation, run `python -m "
                   "SimpleHTTPServer` from the `{}` directory.")
            print(msg.format(self.pydy_directory))

    def remove_static_html(self, force=False):
        """Removes the ``static`` directory from the current working
        directory.

        Parameters
        ----------
        force : boolean, optional, default=False
            If true, no warning is issued before the removal of the
            directory.

        """

        pydy_dir = os.path.join(os.getcwd(), self.pydy_directory)

        if not os.path.exists(pydy_dir):
            print("All Done!")
            return

        if not force:
            msg = ("Are you sure you would like to delete the '{}' "
                   "directory? [y|n]\n")
            ans = raw_input(msg.format(self.pydy_directory))
            if ans == 'y':
                force = True

        if force:
            distutils.dir_util.remove_tree(pydy_dir)
            print("All Done!")
        else:
            print("Aborted!")

    def display(self):
        """Displays the scene in the default web browser."""
        self.create_static_html(overwrite=True, silent=True)
        resource_dir = os.path.join(os.getcwd(), self.pydy_directory)
        server = Server(scene_file=self._scene_json_file,
                        directory=resource_dir)
        server.run_server()

    def _rerun_button_callback(self, btn):
        """Callback for the "Rerun Simulation" button. When executed the
        parameter values are collected from the text input widgets and used
        in a new simulation of the model."""

        btn._dom_classes = ['btn-info', 'active', 'disabled']

        btn.description = 'Rerunning Simulation...'

        original_scene_file = self._scene_json_file
        original_sim_file = self._simulation_json_file
        original_constants = self.system.constants.copy()
        original_initial_conditions = self.system.initial_conditions.copy()
        try:
            self.system.constants = {s: w.value for s, w in
                                     self._constants_text_widgets.items()}
            self.system.initial_conditions = {s: w.value for s, w in
                                              self._initial_conditions_text_widgets.items()}
            pydy_dir = os.path.join(os.getcwd(), self.pydy_directory)
            self._generate_json(directory=pydy_dir)
        except:
            print('Simulation rerun failed, using previous simulation data.')
            # If the simulation fails for any reason we revert everything
            # back to the previous state, including filling the text widgets
            # with the previous values of the constants. The _scene_info and
            # _simulation_data dicts may be in a bad state, but that should
            # be ok, because generate_visualiation_json will have to be run
            # again for anything new to happen.
            self._scene_json_file = original_scene_file
            self._simulation_json_file = original_sim_file
            self.system.constants = original_constants
            self.system.initial_conditions = original_initial_conditions
            self._fill_constants_widgets()
            self._fill_initial_conditions_widgets()

        js_tmp = 'jQuery("#json-input").val("{}/{}");'
        js = js_tmp.format(self.pydy_directory, self._scene_json_file)
        display(Javascript(js))
        display(Javascript('jQuery("#simulation-load").click()'))

        btn._dom_classes = ['btn-info', 'enabled']

        btn.description = self._rerun_button_desc

    def _fill_constants_widgets(self):
        """Fills up the constants widget with the current constants symbols
        and values."""

        for sym, init_val in self._system.constants.items():

            desc = latex(sym, mode='inline')

            text_widget = widgets.FloatText(value=init_val,
                                            description=desc)

            self._constants_text_widgets[sym] = text_widget

    def _fill_initial_conditions_widgets(self):

        for sym in self._system.coordinates + self._system.speeds:

            val = self._system.initial_conditions[sym]

            # TODO : There should be a better way to do this. It would be nice
            # to format the float in a away that always prints compactly and
            # gives enough information even if the number is a very small or
            # very large value.
            desc = latex(sym, mode='inline')
            desc = desc.replace(r'\left ({} \right )'.format(dynamicsymbols._t),
                                '({:1.2f})'.format(self._system.times[0]))

            text_widget = widgets.FloatText(value=val,
                                            description=desc)

            self._initial_conditions_text_widgets[sym] = text_widget

    def _initialize_rerun_button(self):
        """Construct a button for controlling rerunning the simulations."""

        self._rerun_button = widgets.Button()
        self._rerun_button._dom_classes = ['btn-info']
        self._rerun_button_desc = "Rerun Simulation"
        self._rerun_button.description = self._rerun_button_desc
        self._rerun_button.on_click(self._rerun_button_callback)

    def display_ipython(self):
        """Displays the scene using an IPython widget inside an IPython
        notebook cell.

        Notes
        =====
        IPython widgets are only supported by IPython versions >= 3.0.0.

        """

        if IPython is None:
            raise ImportError('IPython is not installed but is required. ' +
                              'Please install IPython >= 3.0 and try again')

        if ipython_less_than_3:
            msg = ('You have IPython {} installed but PyDy only supports '
                   'IPython >= 3.0. Please update IPython and try again.')
            raise ImportError(msg.format(IPython.__version__))

        self.create_static_html(overwrite=True, silent=True)

        # Only create the constants input boxes and the rerun simulation
        # button if the scene was generated with a System.
        if self._system is not None:

            self._initial_conditions_container = widgets.Box()
            self._initial_conditions_container.padding = "10px"
            self._initial_conditions_text_widgets = OrderedDict()
            self._fill_initial_conditions_widgets()
            title = widgets.HTML('<h4 style="margin-left: 5ex;">Initial Conditions</h4>')
            self._initial_conditions_container.children = ((title, ) +
                 tuple(v for v in self._initial_conditions_text_widgets.values()))

            self._constants_container = widgets.Box()
            self._constants_container.padding = "20px"
            self._constants_text_widgets = OrderedDict()
            self._fill_constants_widgets()
            title = widgets.HTML('<h4 style="margin-left: 5ex;">Constants</h4>')
            self._constants_container.children = ((title, ) +
                 tuple(v for v in self._constants_text_widgets.values()))

            self._initialize_rerun_button()

            parameter_input_container = \
                widgets.HBox(children=(self._initial_conditions_container,
                                       self._constants_container))

            display(parameter_input_container)
            display(self._rerun_button)

        ipython_static_url = os.path.relpath(self.pydy_directory, os.getcwd())
        ip_html_file = os.path.join(ipython_static_url, "index_ipython.html")
        with open(ip_html_file, 'r') as html_file:
            html = html_file.read()

        html = html.format(static_url=ipython_static_url,
                           load_url=os.path.join(ipython_static_url,
                                                 self._scene_json_file))

        self._html_widget = widgets.HTML(value=html)
        # NOTE : This overrides the width of the simulation canvas so that it
        # stays within the borders of the IPython notebook.
        self._html_widget._css = [("canvas", "width", "100%"),
                                  ("canvas", "padding-right", "10px")]

        display(self._html_widget)
